//
// Copyright 2002-2016, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include "BrashCommands.h"
#include <cxxtls/stringexpressionparser.h>
#include <cxxtls/file.h>
#include <cxxtls/AutoDestroyingPointer.h>
#include <cxxtls/threadGate.h>
#include <cxxtls/ArithmeticExpressionParser.h>
#include <cxxtls/AutoSaveRestore.h>

#include <signal.h>
#include <fstream>

using namespace cxxtls;
using namespace std;

static std::string apostrophe("'");

static std::string colon(":");
static std::string nl("\r\n");
static std::string semicolon(";");

extern std::string specifiedRCFileName; // instead of .brashrc

Mutex BrashScriptCommandHandler::exitCodeGuard_;
int   BrashScriptCommandHandler::exitCode_;
int   BrashScriptCommandHandler::jobId_;   // also guarded by exitCodeGuard_;


ThreadMonitoredSet<JobDefinition, BrashThreadExitType>  BrashScriptCommandHandler::activeJobs_;


//==================================================================================================
//==================================================================================================
//==================================================================================================

static bool consoleRunning = true;

Mutex  consoleRunningGuard;

bool consoleStillRunning()
{
   // called only statements that have loops in them

   ScopedMutexLock lock(consoleRunningGuard);

   return consoleRunning;
}

namespace cxxtls
{
   bool shuttingDown() { return !consoleStillRunning(); }  // tells tailFile() other functions give up the ghost
}


void setConsoleRunFlag(bool running)
{
   // called only ^C handler and by readConsole

   ScopedMutexLock lock(consoleRunningGuard);

   consoleRunning = running;
}

//==================================================================================================
//==================================================================================================
//==================================================================================================

static set<string> windowsExecutableExtensions;

static int initializeWindowsExecutableExtensions()
{
#ifdef _MSC_VER

    char const *pathext = getenv("PATHEXT");

    if(pathext == 0)
    {
         windowsExecutableExtensions.insert(".EXE");
         windowsExecutableExtensions.insert(".COM");
         windowsExecutableExtensions.insert(".BAT");
         windowsExecutableExtensions.insert(".CMD");
    }
    else
    {
         {
             cxxtls::StrTool::stringlist_t exetypes;
         
             cxxtls::StrTool::Is_Delim  functor(';');
         
             cxxtls::StrTool::parse_words(pathext,
                                  &exetypes,
                                  2000000000U,
                                  functor
                                 );
           
             cxxtls::StrTool::stringlist_t::iterator a = exetypes.begin(),
                                                     b = exetypes.end();
             while(a != b)
             {
                 windowsExecutableExtensions.insert( StrTool::uppercase(*a++) );
             }
         }
        
    }


#endif
    return 1;
}

static int ensureWindowsExecutableExtensions = initializeWindowsExecutableExtensions();

//==================================================================================================
//==================================================================================================
//==================================================================================================

static string expandStringVariableReferences(string const &s, cxxtls::ScriptThread::ThreadContext *context);
   /// assume that s contains $ signs and quotes.  Evaluate and expand as needed to give the
   /// string that is meant.


//==================================================================================================
//==================================================================================================
//==================================================================================================




BrashBuiltinCommandExecutor::
CommandLineArguments::
CommandLineArguments(BrashCommandLine                     * commandLine,
                     cxxtls::ScriptThread::ThreadContext  * executionContext,
                     bool                                   skipArg0,        // for use by command "builtin"
                     bool                                   globbing,
                     bool                                   splitting
                    )
{

    evaluateCommandLine(commandLine, executionContext, args_, skipArg0, globbing, splitting);

    argPointers_.reserve(args_.size()+1);

    CXXTLS_FOREACH(std::string const &cur, args_)
       argPointers_.push_back((char*)(cur.c_str()));

    argPointers_.push_back(0);

    argc = int(args_.size());
    argv = &argPointers_[0];

    if(executionContext->loggingExecution())
    {
        string logMessage("+");



        CXXTLS_FOREACH(std::string const &cur, args_)
        {
            logMessage += ' ';
            logMessage += cur;
        }

        logMessage += nl;

        executionContext->fileHandles_->writeBytes(2, logMessage);
          
    }

}


//==================================================================================================
//==================================================================================================
//==================================================================================================

void
BrashStatementList::
run()
{
    CXXTLS_DEBUG("Begin   BrashStatementList::run(" << std::hex << (long )(this) << std::hex << ")");  
    CXXTLS_DEBUG("Context BrashStatementList::run() = " << (void*)(context_) );

#ifdef DebugBuild
    string debugDescription;    // quick and dirty description of the commands we are executing.  Useful only the debugger
#endif

    
    try
    {
        StatementReturnCode rv;

        try 
        {
#ifdef DebugBuild
           debugDescription = statements_->getDebugDescription();
           CXXTLS_LOG("BrashStatementList: " << (void*)(this) << "  executing " << debugDescription);

           {

               ScriptThread::ManagedFileHandles::FileHandleAttributes stdinInfo  = context_->fileHandles_->getFileAttributes(0);
               ScriptThread::ManagedFileHandles::FileHandleAttributes stdoutInfo = context_->fileHandles_->getFileAttributes(1);


               if(stdinInfo.pipeState_)
               {
                  CXXTLS_LOG("BrashStatementList: " << (void*)(this) << " with PIPESTATE " << (void*)(&*stdinInfo.pipeState_) 
                             << "  stdin ( writeHandle = " << stdinInfo.pipeState_->writeHandle()
                             << ", readHandle = " << stdinInfo.pipeState_->readHandle() 
                             << ")"
                            );
               }

               if(stdoutInfo.pipeState_)
               {
                  CXXTLS_LOG("BrashStatementList: " << (void*)(this) << " with PIPESTATE " << (void*)(&*stdoutInfo.pipeState_)
                             << "  stdout ( writeHandle = " << stdoutInfo.pipeState_->writeHandle() 
                             << ", readHandle = " << stdoutInfo.pipeState_->readHandle()
                             << ")"
                            );
               }

               
           }


#endif
            


            rv = statements_->execute(context_, 0, dynamic_cast<BrashScriptCommandHandler*>(owner_), false ); // need to fix function scope pointer
        }
        catch (CannotOpenTargetFileException &e)
        {
            ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2, string("Brash:  Error opening file:  ") + e.what() + nl);

            rv = ScriptFunctions::SRC_Success;  // don't cancel the interpreter

        }
        catch(ScriptThread::UserIOCancellation &e)
        {
            // treat this as a non-fatal exit

            ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2,string("^C\r\n"));

            rv = ScriptFunctions::SRC_Success;  // don't cancel the interpreter


        }
        catch(ScriptThread::UnableToCreatePipe &e)
        {
            // treat this as a non-fatal exit

            ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2,string("[brash ERROR:  badpipe]\r\n"));

            rv = ScriptFunctions::SRC_Success;  // don't cancel the interpreter


        }
        catch(ScriptThread::PipeWriteFailure &e)
        {
            // treat this as a non-fatal exit

            // produce not log or other status updates, this should be happening a lot in pipes and
            // not be treated as an error

            CXXTLS_LOG("BrashPipeStatement: " << (void*)(this) << ", ignoring PIPE WRITE FAILURE");

        }
        catch(ProgramStructureException &e)
        {
            // do nothing, assume the thrower has printed the needful error message
        }
        catch(...)
        {
            throw;  // let outer handler deal with it
        }

        try
        {
            if(startedByConsole_)
            {
                
                BrashScriptCommandHandler::removeJob( jobId_, 
                                                      BrashThreadExitType( context_->variables().errorCode(), 
                                                                           context_->fileHandles_->currentDirectory(),
                                                                           context_->variables().arguments(),
                                                                           context_->variables().lastParmOfLastCommand(),
                                                                           context_->variables().find("DIRSTACK"),
                                                                           context_->variables().find("IFS"),
                                                                           context_->variables().globbing()
                                                                         ) 
                                                    );
            }
            else
            {
                // normally, we are not started by the console.
            
                BrashScriptCommandHandler::removeJob( jobId_, 
                                                      BrashThreadExitType( context_->variables().errorCode(), 
                                                                           std::string(), 
                                                                           std::vector< std::string >(),
                                                                           std::string(),
                                                                           std::string(),
                                                                           std::string(),
                                                                           true
                                                                        ) 
                                                    );
            }
        }
        catch(...)
        {
            CXXTLS_DEBUG("Brash Cannot remove job " << jobId_ << " from the queue");
            throw;
        }

        if(rv == ScriptFunctions::SRC_Break ||
           rv == ScriptFunctions::SRC_Continue
          )
        {
            context_->fileHandles_->writeBytes(2,string("break and continue statements are only allowed in while and for statements\r\n"));
            context_->variables().setErrorCode(1);  // an error occurred
        }

        if(    !suppressExit_ 
           &&
               rv == ScriptFunctions::SRC_Return
          )
        {
            // when executing external scripts, we do not want to handle exit statements as a directive
            // to shutdown the whole brash executable

            BrashScriptCommandHandler::shutdownJobs();

            shutdown();
        }

    }
    catch(std::exception &e)
    {
        ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2,string("brash unhandled exception: "));
        ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2,e.what());
        ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2,"\r\n", 1);

        BrashScriptCommandHandler::shutdownJobs();

        shutdown();
    }
    catch(...)
    {
        ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2,string("brash unhandled exception in BrashStatementList::run()"));

        BrashScriptCommandHandler::shutdownJobs();

        shutdown();
    }

    CXXTLS_DEBUG("End BrashStatementList::run(" << std::hex << (long )(this) << std::hex << ")");  

}

//==================================================================================================
//==================================================================================================
//==================================================================================================

static void drainStdin(cxxtls::ScriptThread::ThreadContext  * executionContext)
{
    ScriptThread::ManagedFileHandles::iterator stdinInfo = executionContext->fileHandles_->find(0);

    if (stdinInfo != executionContext->fileHandles_->end())
    {
        if (stdinInfo->second.pipeState_)
        {
            if (!stdinInfo->second.pipeState_->terminated() && stdinInfo->second.terminatePipe_)
            {
                stdinInfo->second.pipeState_->cancelWrite();
            }
        }


        if (stdinInfo->second.pipeState_
            && stdinInfo->second.pipeState_->writeHandle()
            && !stdinInfo->second.isConsole()
            )
        {
            // now read and discard any remaining data in the stdin handle. unless we are in a while loop or for loop

            char buffer[8096];

            DWORD inputBytes = 0;

            while (ReadFile(
                stdinInfo->second.msHandle_,
                buffer,
                sizeof(buffer),
                &inputBytes,
                0
                )
                );
        }

    }

}


void
BrashPipeMember::
run()
{
    // this function is running in its own theread
    CXXTLS_DEBUG("ThreadContext      BrashPipeMember::run() = " << (void*)(context_) );
    CXXTLS_DEBUG("ExecutionContext   BrashPipeMember::run() = " << (void*)(&context_->variables()) );
    CXXTLS_DEBUG("ExecutionVariables BrashPipeMember::run() = " << (void*)(&context_->variables().variables()) );

#ifdef DebugBuild
    string debugDescription;    // quick and dirty description of the commands we are executing.  Useful only the debugger
#endif

    try
    {
        StatementReturnCode  rv = ScriptFunctions::SRC_Success;
        

        try 
        {
#ifdef DebugBuild
            debugDescription = statement_->getDebugDescription();
            CXXTLS_LOG("BrashPipeMember: " << (void*)(this) << ", begins  executing " << debugDescription);

           {

               ScriptThread::ManagedFileHandles::FileHandleAttributes stdinInfo  = context_->fileHandles_->getFileAttributes(0);
               ScriptThread::ManagedFileHandles::FileHandleAttributes stdoutInfo = context_->fileHandles_->getFileAttributes(1);


               if(stdinInfo.pipeState_)
               {
                  CXXTLS_LOG("BrashPipeMember: " << (void*)(this) << " with PIPESTATE " << (void*)(&*stdinInfo.pipeState_)
                             << "  stdin ( writeHandle = " << stdinInfo.pipeState_->writeHandle()
                             << ", readHandle = " << stdinInfo.pipeState_->readHandle()
                             << ")"
                            );
               }

               if(stdoutInfo.pipeState_)
               {
                  CXXTLS_LOG("BrashPipeMember: " << (void*)(this) << " with PIPESTATE " << (void*)(&*stdoutInfo.pipeState_)
                             << "  stdout ( writeHandle = " << stdoutInfo.pipeState_->writeHandle()
                             << ", readHandle = " << stdoutInfo.pipeState_->readHandle()
                             << ")"
                            );
               }

               
           }

#endif

            try
            {
                rv = statement_->execute(context_, 0, dynamic_cast<BrashScriptCommandHandler*>(owner_) ); // need to fix function scope pointer
            }
            catch(...)
            {
                drainStdin(context_);
                throw;
            }


            // we have run a command, or commands in the pipe
            //
            // at this point, we must be concerned with left over data hanging around in pipes.  The Windows WriteFile
            // function will hang until someone reads from a pipe.  It does not recognize that the downstream side of the
            // pipe is closed.  To get Windows to break out of this hung write operation, you need to call CancelIoEx().
            // Not just CancelIo(), but the Ex variant.  The pipeState_->cancelWrite() call does that.

            if (terminateInputPipe_)
                drainStdin(context_);

        }
        catch(ScriptThread::UnableToCreatePipe &e)
        {
            // treat this as a non-fatal exit

            ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2,string("[brash ERROR:  badpipe]\r\n"));


        }
        catch(ScriptThread::UserIOCancellation &e)
        {
            // treat this as a non-fatal exit

            ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2,string("^C\r\n"));


        }
        catch (ScriptThread::PipeWriteFailure &e)
        {
            // treat this as a non-fatal exit

            // pipe write failure exceptions are used to terminate script fragments that are writing to pipes when the downstream
            // script fragment terminates early.
            //
            // These exceptions do not signal an error but rather the end of processing, so produce no error messages.

            CXXTLS_LOG("BrashPipeMember: " << (void*)(this) << ", ignoring PIPE WRITE FAILURE");

            context_->variables().setErrorCode(1);

        }
        catch(ProgramStructureException &e)
        {
            // do nothing, assume the thrower has printed the needful error message

        }
        catch(RedirectionException const &e)
        {
            // normally, we are not started by the console.

            context_->variables().setErrorCode(1);

            string message="brash ERROR:  bad pipe member redirection in statement:  ";



            message += statement_->getDebugDescription();

            ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2,message + nl);

    

        }
        catch(...)
        {
            // normally, we are not started by the console.


            if(terminateInputPipe_)
                context_->fileHandles_->terminatePipe(0);

            BrashScriptCommandHandler::removeJob( jobId_, 
                                                  BrashThreadExitType( context_->variables().errorCode(), 
                                                                       std::string(), 
                                                                       std::vector< std::string >(),
                                                                       std::string(),
                                                                       std::string(),
                                                                       std::string(),
                                                                       true
                                                                    ) 
                                                );

            context_->variables().setErrorCode(1);


            throw;  // let outer handler deal with hit
        }

        if(terminateInputPipe_)
            context_->fileHandles_->terminatePipe(0);

        {
            // normally, we are not started by the console.

            BrashScriptCommandHandler::removeJob( jobId_, 
                                                  BrashThreadExitType( context_->variables().errorCode(), 
                                                                       std::string(), 
                                                                       std::vector< std::string >(),
                                                                       std::string(),
                                                                       std::string(),
                                                                       std::string(),
                                                                       true
                                                                    ) 
                                                );
        }


#if 0

        // you can't return or exit from a pipe statement.  Just ignore attempts to do so.

        if(rv == ScriptFunctions::SRC_Return)
        {
            BrashScriptCommandHandler::shutdownJobs();

            shutdown();
        }
#endif

        if(rv == ScriptFunctions::SRC_Break ||
           rv == ScriptFunctions::SRC_Continue
          )
        {
            context_->fileHandles_->writeBytes(2,string("break and continue statements are only allowed in while and for statements\r\n"));
            context_->variables().setErrorCode(1);  // an error occurred
        }


    }
    catch(std::exception &e)
    {
        drainStdin(context_);
        if(terminateInputPipe_)
            context_->fileHandles_->terminatePipe(0);

        ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2,string("brash unhandled exception: "));
        ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2,e.what());
        ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2,nl);

        finishLine_.signal();  // my pipe statement won't terminate for the shutdown to complete

        BrashScriptCommandHandler::shutdownJobs();

        shutdown();
    }

}


//==================================================================================================
//==================================================================================================
//==================================================================================================


StatementReturnCode 
BrashNoOpStatement::
execute(  cxxtls::ScriptThread::ThreadContext          *executionContext,
          cxxtls::ScriptFunctions::FunctionScope       *functionScope,
          cxxtls::ScriptFunctions::JobController       *jobController,
          bool                                          skipArg0
       ) const
{
       executionContext->fileHandles_->writeBytes(1,std::string("BrashNoOp Statement executes:\r\n"));

       cxxtls::StringExpr::ExecutionContext   &context = executionContext->variables();
       cxxtls::StringExpr::ExecutionVariables &env = context.variables();

       executionContext->fileHandles_->writeBytes(1, "Arguments: \r\n");

       CXXTLS_FOREACH(std::string const &cur, context.arguments())
       {
           executionContext->fileHandles_->writeBytes(1, "          ");
           executionContext->fileHandles_->writeBytes(1, cur);
           executionContext->fileHandles_->writeBytes(1, nl);
       }


       cxxtls::StringExpr::ExecutionVariables::const_iterator f = env.begin(), l = env.end();

       while(f != l)
       {
           cxxtls::StringExpr::ExecutionVariables::VariablesMap::value_type const &cur = *f++;

           executionContext->fileHandles_->writeBytes(1, cur.first);
           executionContext->fileHandles_->writeBytes(1, std::string("="));
           executionContext->fileHandles_->writeBytes(1, cur.second);
           executionContext->fileHandles_->writeBytes(1, nl);

       }

       return ScriptFunctions::SRC_Success;


}

//==================================================================================================
//==================================================================================================
//==================================================================================================

string BrashCommandLineStatement::description() const
{
   StringExpression * arg0 = *command_->arguments_->begin();

   string name;

   if(arg0->isSimple(name) )
   {
      return name;
   }

   return "unspecified command line statement";
}

//==================================================================================================
//==================================================================================================
//==================================================================================================

StatementReturnCode 
BrashCommandLineStatement::
execute(  cxxtls::ScriptThread::ThreadContext          *executionContext,
          cxxtls::ScriptFunctions::FunctionScope       *functionScope,
          cxxtls::ScriptFunctions::JobController       *jobController,
          bool                                          skipArg0
       ) const
{
    StatementReturnCode rv;

    try
    {

     rv=   commandExecutor_->execute(  command_.get(),
                                       executionContext,
                                       functionScope,
                                       jobController,
                                       skipArg0
                                    );

    }
#if 0
    catch(RedirectionException const &e)
    {
       throw; // ignore it here
    }
#endif
    catch(ScriptThread::PipeWriteFailure const &e)
    {
      throw;  // ignore it here
    }
    catch(std::exception const &e)
    {
        executionContext->fileHandles_->writeBytes(2, string("brash error: exception (") + string(e.what()) + string(" at ")
                                                      + command_->sourceFile_ + colon
                                                      + lexical_cast<std::string>(command_->sourceLine_) + nl
                                                  );
       rv =  ScriptFunctions::SRC_Success;
    }
    
    return rv;
}


//==================================================================================================
//==================================================================================================
//==================================================================================================

DECLARE_CXXTLS_EXCEPTION(UnexpectedArrayIndexInCommandLineVariableSuperseed);

BrashBuiltinCommandExecutor::
WorkingHandlesAndEnvironmentPointer::
WorkingHandlesAndEnvironmentPointer(BrashCommandLine*           commandLine,
                          cxxtls::ScriptThread::ThreadContext  *executionContext_in
                         )
{
    originalContext_ = executionContext_in;  // does not change

    if(!commandLine)
    {
        throw BrashCommandLineStatement::EmptyCommandException();
    }

    BrashCommandLine::REDIRECTION_LIST const &redirs = *commandLine->redirections_;

    workingContext_  = executionContext_in;


    if(!redirs.empty() || (commandLine->superseeds_.get() && !commandLine->superseeds_->empty() ) )
    {
        // there are file redirections, so superseed the file workingFileHandles with a modified
        // copy



        clonedContext_.reset( executionContext_in->cloneForSameJob(workingContext_->variables().scriptName(), workingContext_->variables().arguments()) );

        if (!clonedContext_.get())
        {
            throw CannotDupBadFileException("Bad Handle", "Parent shell environment corrupted");
        }

        workingContext_ = clonedContext_.get();

        if( (commandLine->superseeds_.get() && !commandLine->superseeds_->empty() ) )
        {
           // add the command line's superseeds to any already in the environment

           typedef pair<const std::string, BrashCommandLine::ValueAndIndex> SUPERSEED;

           CXXTLS_FOREACH(SUPERSEED &cur, *commandLine->superseeds_)
           {
               StringArray output;

               if(cur.second.index_)
               {
                  throw UnexpectedArrayIndexInCommandLineVariableSuperseed();
               }

               cur.second.value_->execute(&executionContext_in->variables(), output);

               workingContext_->variables().superseed(cur.first, output.combine());   // we are not setting the variable in the environment, we are copying it into the superseeds thereof
           }
        }

        workingContext_->variables().setErrorCode( executionContext_in->variables().errorCode() );



        try
        {
            RedirType::redirectFileHandles(workingContext_, &redirs );

        }
        catch(...)
        {

            workingContext_->variables().setErrorCode(1); // make sure caller gets an error
            throw;
        }

    }

   
}


//==================================================================================================
//==================================================================================================
//==================================================================================================


void
BrashBuiltinCommandExecutor::
evaluateCommandLine(BrashCommandLine                     * commandLine,
                    cxxtls::ScriptThread::ThreadContext  * executionContext,
                    std::vector<std::string>             & argsOut,
                    bool                                   skipArg0,
                    bool                                   globbing, 
                    bool                                   splitting
                   ) 
{
         std::list<cxxtls::StringExpression*> const &args = *commandLine->arguments_;

         static string blank(" ");

         argsOut.clear();

         argsOut.reserve(args.size());

         int count = 0;

         CXXTLS_FOREACH(StringExpression const *cur, args)
         {
             ++count;

             if(skipArg0 && count == 1)
             {
                continue;  // this only happens if the builtin command is operating on this commandLine
             }


             StringArray output;

             #ifdef DebugBuild

             string expressionNode = cur->stringify();  

             #endif

             cur->executeHelper(&executionContext->variables(), output, globbing, splitting);

             for (size_t i = 0; i < output.size(); ++i)
             {
                 string const &cur = output[i];
                 argsOut.push_back(cur);
             }
         }

         if(argsOut.size() > 1)
             executionContext->variables().setLastParmOfLastCommand(argsOut.back());
         else
              executionContext->variables().setLastParmOfLastCommand(std::string());
}





//==================================================================================================
//==================================================================================================
//==================================================================================================

BrashBuiltinCommandExecutor const *findBuiltin(string const &name)
{
   map<string, BrashBuiltinCommandExecutor*>::const_iterator where = builtinCommands.find(name);

   if(where == builtinCommands.end())
      return 0;


   return where->second;
}

//==================================================================================================
//==================================================================================================
//==================================================================================================

void
BrashCommandLineStatement::
detectExecutionClass()
{
    if(command_.get() == 0)
       throw EmptyCommandException();

    if(command_->arguments_->empty())
    {
        throw EmptyCommandException();
    }

    StringExpression const *command = *command_->arguments_->begin();

    if(!command)
    {
        throw EmptyCommandException();
    }

    if(command->empty())
    {
        throw EmptyCommandException();
    }

    string commandName;

    commandExecutor_ = &brashBuiltinExecuteExternalCommand;

    if(!command->isSimple(commandName))
    {
       // we are done
    }
    else
    {

       BrashBuiltinCommandExecutor const *p = findBuiltin(commandName);

       if(p)
       {
           commandExecutor_ = p;
       }

    }



}

//==================================================================================================
//==================================================================================================
//==================================================================================================


BrashExternalCommandExecutor brashBuiltinExecuteExternalCommand;

string 
BrashExternalCommandExecutor::
abstract() const
{
   return "Un-named builtin command executor that runs programs and functions";
}

std::string parameterQuotify(std::string const &r, bool forceQuotes=false)
    // put quotes around a string if forceQuotes is true, or if the string contains a character that
    // windows transmogrifies.
{
    char const *needsQuoting = " \t\r\n()^%|&,;=";   // characters that should be quoted when using them in a commad line call to CallProcess()

    if(   !forceQuotes 
       && (r.find_first_of(needsQuoting) >= r.size())
      )
    {
        return r;
    }


    std::string rv;

    rv += '"';

    CXXTLS_FOREACH(char c, r)
    {
       if(c == '"')
       {
          rv += '\\';
       }

       rv += c;
    }

    rv += '"';

    return rv;
}

StatementReturnCode
BrashExternalCommandExecutor::
execute(BrashCommandLine*                             commandLine,
        cxxtls::ScriptThread::ThreadContext          *executionContext_in,
        cxxtls::ScriptFunctions::FunctionScope       *functionScope,
        cxxtls::ScriptFunctions::JobController       *jobController,
        bool                                          skipArg0      //  for the command, "builtin"
       ) const
{

    {

        bool isPipeMiddle = false;   // the flag that tells callProcess if we are in the middle of a pipe
    
        if(executionContext_in)
        {
           // compute isPipeMiddle
    
           ScriptThread::ManagedFileHandles::iterator stdinInfo = executionContext_in->fileHandles_->find(0);
    
           if(stdinInfo != executionContext_in->fileHandles_->end())
           {
              // we have a stdin file handle
    
              ScriptThread::ManagedFileHandles::FileHandleAttributes &info = stdinInfo->second;
    
              if(info.pipeState_)
              {
                  if(info.pipeState_->writeHandle())
                     isPipeMiddle = true;  // stdin is a pipe and there is a writer for it
    
              }
           }
        }
    
        
        if(!commandLine)
        {
            throw BrashCommandLineStatement::EmptyCommandException();
        }
    
        //  WorkingFileHandlesPointer  workingFileHandles(commandLine, executionContext);  // treat this a pointer to the current ManagedFileHandles object
    
        WorkingHandlesAndEnvironmentPointer  workingContext(commandLine, executionContext_in);
    
        cxxtls::ScriptThread::ThreadContext        *executionContext = workingContext.executionContext();
        cxxtls::ScriptThread::ManagedFileHandles   *workingFileHandles = workingContext.fileHandles();
    
    
        bool thereAreSuperseeds = (executionContext != executionContext_in);   // we'll need this info when we invoke callProcess
        CommandLineArguments arguments(commandLine, executionContext, skipArg0);
    
    
        if (arguments.args_.empty())
        {
            workingFileHandles->writeBytes(2, string("brash Error executing empty command line at ") +  commandLine->sourceFile_ + colon + lexical_cast<string>(commandLine->sourceLine_) + nl);
            return ScriptFunctions::SRC_Success;  
        }
    
        string cmdline;
    
        cmdline.reserve(arguments.args_.size() *40 );  // guess average of 40 characters as a starting size
    
        bool isFunctionCommand = false; // this is not a shell function.
    
        StringExpr::ExecutionVariables::FunctionInfo funcHandle = workingContext.executionContext()->variables().findFunction(arguments.args_[0]);
    
        if(funcHandle.statement_)
        {
           isFunctionCommand = true;
        }
    
        bool pathIsKnown = false;
    
        bool foundBrashScript = false;
        string brashScriptFound;                  // only used if foundBrashScript is true
    
        if(!isFunctionCommand)
        {
            pair<string,int> execType = workingContext.executionContext()->findInExecHash(arguments.args_[0],0);   // create as native if not already found
    
            if (execType.second < 1
                && execType.first.empty()
                )
            {
                // if we couldn't find the command path, and if the name has not . characters in it, try looking for command.brash.
    
                string commandName = arguments.args_[0];
    
                if (commandName.find_first_of('.' >= commandName.size()))
                {
                   // the command could not be found, and the command name does not contain a '.' character -- try sticking .brash on the end of the name
    
                    commandName += ".brash";
    
                    string PATH=workingContext.executionContext()->variables().find("PATH");
    
                    string found = FileName::find_file_in_path(commandName, PATH); 
    
                    if(!found.empty())
                    {
                        execType.first = found;
                        execType.second= 1;
    
    #ifdef _MSC_VER
                        // on windows, normalize the / to backslash
                        string tmp;
    
                        CXXTLS_FOREACH(char c, execType.first)
                        {
                           if(c == '/')
                              c = '\\';
    
                           tmp += c;
                        }
    
                        execType.first = tmp; 
    #endif
    
                        workingContext.executionContext()->addToExecHash(arguments.args_[0], execType.first, 1);
    
                    }
                }
            }
            
            if(execType.second < 1)
            {
                cmdline += execType.first;  // native binary perhaps, or maybe a script -- don't know yet
    
                FileName tmpCmdName = execType.first;
    
    #ifdef _MSC_VER
                if(tmpCmdName.exists())
                {
                    // there is a file by the name specified
    
                    string extension = tmpCmdName.extension();
    
                    if(windowsExecutableExtensions.find(extension) == windowsExecutableExtensions.end())
                    {
                        // this file is not a native binary
    
                        ifstream file(tmpCmdName.c_str());
    
                        string firstLine;
    
                        std::getline(file, firstLine);
    
                        if(firstLine.size() > 2)
                        {
                           if(firstLine[0] == '#' && firstLine[1] == '!')
                           {
                               // we have a script
    
                               foundBrashScript = true;
    
                               brashScriptFound = execType.first;
    
                           }
                        }
    
    
                    }
                }
    #else
    #endif
    
            }
            else
            if(execType.second == 1)
            {
               // a brash script
                cmdline += execType.first;
    
                foundBrashScript = true;
    
                brashScriptFound = execType.first;
            }
    
        }
    
        if (cmdline.empty())
            cmdline = arguments.args_[0];
        else
            pathIsKnown = true;
    
    #ifdef _MSC_VER
        bool isCmd=false;
    
        if(!arguments.args_.empty() )
        {
           if(arguments.args_[0] == "cmd")
           {
              isCmd = true;
           }
        }
    #endif
    
        if(arguments.argc > 1)
           cmdline += ' ';
    
    #ifdef _MSC_VER
        if(isCmd 
           && arguments.args_.size() > 1
           && !arguments.args_[1].empty()
           && arguments.args_[1][0] == '/'
          )
        {
           // wrap up all the NON-OPTION arguments in a single block for cmd.exe.  For example
           //
           //    cmd /c "dir "a" "b" "c" "d""
           //
           //           ^-------------------^
           //
    
           int i = 1;   // argument index
    
           // copy all the options to "cmd" itself without putting quotes on them
    
           while(    i < arguments.argc
                 && !arguments.args_[i].empty()
                 && arguments.args_[i][0] == '/'
                )
           {
               cmdline += arguments.args_[i];
               cmdline += ' ';
    
               ++i;
           }
    
    
           // due to bugs in the way cmd.exe interprets the first parameter (I say bugs because it
           // is very inconvenient for no apparent reason) I must not put quotes on the command name --
           // unless of course it has spaces in it -- in which case, it is not a builtin command
           // but a program name.  The copy command is builtin, so you can't put quotes around copy.
    
           if(i < arguments.argc)
           {
               if(arguments.args_[i].find_first_of(" ") < arguments.args_.size())
               {
                  // the command name has a space in it
    
                  cmdline += parameterQuotify(arguments.args_[i], true);
               }
               else
                  cmdline += arguments.args_[i];
    
               cmdline += ' ';
    
               ++i;

           }
    
           // now, avoid putting quotes on the /options to the program
    
           while(   i < arguments.argc
                 && !arguments.args_[i].empty()
                 && arguments.args_[i][0] == '/'
                )
           {
               cmdline += arguments.args_[i];
               cmdline += ' ';
    
               ++i;
           }
    
    
           if(i < arguments.argc)
           {
               
               for(; i < arguments.argc; ++i)
               {
                  cmdline += parameterQuotify(arguments.args_[i]);
               
                  if(i != arguments.argc-1)
                     cmdline += ' ';
               }
               
           }
    
        }
        else
    #endif
        {   // keep with else clause in the above #ifdef block!
            
            bool forceQuotes = false;
            
            if(arguments.args_[0] == "find")
               forceQuotes = true;
            
            for(int i = 1; i < arguments.argc; ++i)
            {
               cmdline += parameterQuotify(arguments.args_[i], forceQuotes);
            
               if(i != arguments.argc-1)
                  cmdline += ' ';
            }
        }
    
        string officialCurrentDirectory = executionContext->variables().find("PWD");
    
        if(!officialCurrentDirectory.empty())
            workingFileHandles->setCurrentDirectory(officialCurrentDirectory);
    
    
        if(isFunctionCommand)
        {
             if(workingContext.executionContext()->loggingExecution())
             {
                string logMessage("+ function");
    
                CXXTLS_FOREACH(string const &cur, arguments.args_)
                {
                    logMessage += ' ';
                    logMessage += cur;
                }
    
                logMessage += nl;
    
                workingFileHandles->writeBytes(2, logMessage);
             }
    
    
             vector<string> savedArgs = workingContext.executionContext()->variables().arguments();
             string         curScript = workingContext.executionContext()->variables().scriptName();
    
             ScriptFunctions::StatementList  tmpStatement;
    
             tmpStatement.push_back(funcHandle.statement_);
    
             workingContext.executionContext()->variables().setArguments(arguments.args_);
             workingContext.executionContext()->variables().setScriptName(arguments.args_[0]);
    
             StatementReturnCode rc = ScriptFunctions::SRC_Success;
    
             try
             {
                string savedIndex = workingContext.executionContext()->variables().find("OPTIND");
                string savedOffset = workingContext.executionContext()->variables().find("OPTOFF");
                string lastParm = workingContext.executionContext()->variables().lastParmOfLastCommand();
    
                workingContext.executionContext()->variables().assign("OPTIND", "1");
                workingContext.executionContext()->variables().assign("OPTOFF", "0");
    
                 rc = 
                 tmpStatement.execute(workingContext.executionContext(),
                                      functionScope,
                                      jobController,
                                      skipArg0
                                     );
    
                workingContext.executionContext()->variables().assign("OPTIND", savedIndex);
                workingContext.executionContext()->variables().assign("OPTOFF", savedOffset);
    
                workingContext.executionContext()->variables().setArguments(savedArgs);
                workingContext.executionContext()->variables().setLastParmOfLastCommand(lastParm);
                workingContext.executionContext()->variables().setScriptName(curScript);
    
    
             }
             catch(ProgramStructureException &e)
             {
                 // program structure exceptions occur when an [-statement or an ((-statement has been
                 // parsed successfully but contains a syntax error that is only detectable at run-time.
                 //
                 // for example.  [ 1 == 2 ]
                 //
                 // This is a syntax error because the proper syntax is [ 1 = 2 ]
                 //
                 
                 workingFileHandles->writeBytes(2, string("Exception in function call:  "));
    
                 for(size_t i = 0; i < arguments.args_.size(); ++i)
                 {
                     workingFileHandles->writeBytes(2, arguments.args_[i] + string(" "));
                 }
    
    
                 workingFileHandles->writeBytes(2, nl + funcHandle.file_ + colon + lexical_cast<string>(funcHandle.line_) + 
                                                        string(": Error in the definition of function ") + arguments.args_[0] + nl);
    
    
    
                 throw;  // rethrow the exception
             }
    
             if(!arguments.args_.empty())
             {
                 workingContext.executionContext()->variables().setLastParmOfLastCommand(arguments.args_.back());
             }
    
    
             if(rc == ScriptFunctions::SRC_Return)
                rc = ScriptFunctions::SRC_Success;
    
             return rc;
    
        }
        else
        {
            int rc;
    
            if(!pathIsKnown || foundBrashScript)
            {
    
                std::pair<std::string,bool> tmp;
                
                if(!foundBrashScript)
                     tmp = findBrashExecutableInPath(arguments.args_[0], workingContext.executionContext()->variables().find("PATH"));
                else
                {
                    tmp.second = 1;
                    tmp.first =   brashScriptFound;
                }
    
                if(workingContext.executionContext()->loggingExecution())
                {
                   string logMessage;
    
                   if(tmp.second)
                      logMessage = "+ external script: ";
                   else
                      logMessage = "+ external program: ";
                
                   CXXTLS_FOREACH(string const &cur, arguments.args_)
                   {
                       logMessage += ' ';
                       logMessage += cur;
                   }
                
                   logMessage += nl;
                
                   workingFileHandles->writeBytes(2, logMessage);
                }
    
    
                if(!tmp.second)
                {
                    if(thereAreSuperseeds)
                    {
                        // we have a local copy of the environment
    
                        cxxtls::StringExpr::ExecutionContext   &context = executionContext->variables();
                        
                        const cxxtls::StringExpr::ExecutionVariables &supers = context.superseeds();
    
                        StringExpr::ExecutionVariables  tmp = workingContext.executionContext()->variables().variables();
    
                        cxxtls::StringExpr::ExecutionVariables::const_iterator f = supers.begin(), l = supers.end();
                 
                        while(f != l)
                        {
                            cxxtls::StringExpr::ExecutionVariables::VariablesMap::value_type const &cur = *f++;
    
                            string const &name = cur.first;
                            string const &value= cur.second;
    
                            tmp.set(name,value);
    
                            tmp.setExportFlag(name, true);
    
                        }
    
                        rc = workingFileHandles->callProcess(cmdline, tmp, isPipeMiddle); // normal executable, not a script
    
    
                    }
                    else
                    {
                        rc = workingFileHandles->callProcess(cmdline, workingContext.executionContext()->variables().variables(), isPipeMiddle); // normal executable, not a script
                    }
    
                   if(!arguments.args_.empty())
                   {
                       workingContext.executionContext()->variables().setLastParmOfLastCommand(arguments.args_.back());
                   }
    
                }
                else
                {
                   // run the script
    
                    workingContext.executionContext()->variables().variables().addToExecHash(arguments.args_[0], tmp.first, 1);  // add to hash table as brash script
    
                   // workingFileHandles->writeBytes(2, string("Ignored Script:  '") + tmp.first + nl);
    
                   StringExpr::ExecutionVariables::ArgsVector savedArgs, newArgs ;
                   
                   newArgs.push_back(tmp.first);  // arg0
    
                   for(size_t i = 1; i < arguments.args_.size(); ++i)
                       newArgs.push_back(arguments.args_[i]);
    
                   string myscriptname = tmp.first;
    
                   workingContext.executionContext()->variables().swapArguments(savedArgs, myscriptname);
    
                   AutoDestroyingPointer<ScriptThread::ThreadContext>   childContext(workingContext.executionContext()->cloneForNewJob());
    
                   brashReadAndExecuteInContext(tmp.first, childContext.get(), newArgs, true);
                      /// we are passing true for suppressExit, which is normally false.  This parameter
                      /// has the effect of ignoring exit statements in external scripts
    
                   workingContext.executionContext()->variables().swapArguments(savedArgs, myscriptname);
    
                   workingContext.variables().setErrorCode(childContext->variables().errorCode());  // save the child's exit code
    
                   if(!arguments.args_.empty())
                   {
                       workingContext.executionContext()->variables().setLastParmOfLastCommand(arguments.args_.back());
                   }
    
    
                   return ScriptFunctions::SRC_Success;
    
    
                }
    
            }
            else
            {
                if (thereAreSuperseeds)
                {
                    // we have a local copy of the environment
    
                    cxxtls::StringExpr::ExecutionContext   &context = executionContext->variables();
    
                    const cxxtls::StringExpr::ExecutionVariables &supers = context.superseeds();
    
                    StringExpr::ExecutionVariables  tmp = workingContext.executionContext()->variables().variables();
    
                    cxxtls::StringExpr::ExecutionVariables::const_iterator f = supers.begin(), l = supers.end();
    
                    while (f != l)
                    {
                        cxxtls::StringExpr::ExecutionVariables::VariablesMap::value_type const &cur = *f++;
    
                        string const &name = cur.first;
                        string const &value = cur.second;
    
                        tmp.set(name, value);
    
                        tmp.setExportFlag(name, true);
    
                    }
    
                    rc = workingFileHandles->callProcess(cmdline, tmp, isPipeMiddle); // normal executable, not a script
    
    
                }
                else
                {
                    rc = workingFileHandles->callProcess(cmdline, workingContext.executionContext()->variables().variables(), isPipeMiddle); // normal executable, not a script
                }
            }
    
            if(!arguments.args_.empty())
            {
                workingContext.executionContext()->variables().setLastParmOfLastCommand(arguments.args_.back());
            }
    
            
            executionContext->variables().setErrorCode(rc);
            
            if(rc < 0)
            {
                workingFileHandles->writeBytes(2, string("Unable to execute program or script:  '") + cmdline + apostrophe + nl);
            }
        }
    
        return ScriptFunctions::SRC_Success;
    }


} // execute




//==================================================================================================
//==================================================================================================
//==================================================================================================

extern string interpreterExecutableName;

static string computePrompt(string const &prompt, string const &currentDirectory, int lineNumber, string const &userName, string const &hostName)
{
   string rv;

   string::const_iterator f = prompt.begin(),
                          l = prompt.end();

   string shortHost;

   size_t dotOffset = hostName.find_first_of('.');

   if(dotOffset)
      shortHost = hostName.substr(0, dotOffset);
   else
      shortHost = hostName;


   while(f != l)
   {

       char c = *f++;

       if(c == '\\')
       {
           if(f == l)
               rv += '\\' ;
           else
           switch(c = *f++)
           {
              case 'w':  rv += currentDirectory; break;
              case '#':  rv += lexical_cast<string>(lineNumber); break;
              case 'u':  rv += userName; break;
              case 'h':  rv += shortHost; break;
              case 'H':  rv += hostName;  break;
              case 'n':  rv += nl;  break;

              case 's':  
                         if(!interpreterExecutableName.empty())
                         {
                             size_t lastSlash = interpreterExecutableName.find_last_of("\\/");
                             
                             if(lastSlash < interpreterExecutableName.size()-1)
                             {
                                 rv += interpreterExecutableName.substr(lastSlash+1);
                             }
                         }
                         break;
              case 'W':
                         if(!currentDirectory.empty())
                         {
                             size_t lastSlash = currentDirectory.find_last_of("\\/");
                             
                             if(lastSlash < currentDirectory.size()-1)
                             {
                                 rv += currentDirectory.substr(lastSlash+1);
                             }
                         }
                         break;
              case '\\':
                  rv += c;
                  break;

              default:
                  rv += '\\';
                  rv += c;
                  break;
           }
       }
       else
          rv += c;
   }


   return rv;
}


static void ignoreSignal(int)
{
    setConsoleRunFlag(false);     // stop running until the console re-enables
    signal(SIGINT,ignoreSignal);
}

extern bool noReadBrashrc;

void
ReadConsoleCommand::
run()
{

try
{
    signal(SIGINT,ignoreSignal);    // sets the signal handler

    setConsoleRunFlag(true);

    // string prompt("brash>");

    ScriptThread::ManagedFileHandles::FileHandleAttributes h = ScriptThread::ManagedFileHandles::mainStdHandles->getFileAttributes(0);

    string prompt("brash> ");

#ifdef _MSC_VER
    ScriptLineStream console(h.msHandle_,false, prompt);
#else
    ScriptLineStream console(h.unixHandle_,false, prompt);
#endif

    console.recordHistory();  


    int exitCode = 0;

    string  currentDirectory;

    currentDirectory = ScriptThread::ManagedFileHandles::mainStdHandles->currentDirectory();

    console.setWorkingDirectory(currentDirectory); // for tab-completions


    string defaultPrompt="\\w: ";

    cxxtls::StringExpr::ExecutionContext::ArgsVector commandLine;  // used

    StringExpr::ExecutionVariables *workingEnvironment = &env_;




    AutoDestroyingPointer<cxxtls::ScriptThread::ThreadContext>   myContext(ScriptThread::ThreadContext::create("[Brash]consoleReader",
                                                                                                               commandLine,
                                                                                                               workingEnvironment,
                                                                                                               ScriptThread::ManagedFileHandles::mainStdHandles->clone()
                                                                                                              )
                                                                          );


    string workingPrompt = workingEnvironment->get("PS1");

    if(workingPrompt.empty())
       workingPrompt = defaultPrompt;

    console.changePrompt(computePrompt(workingPrompt, 
                                       currentDirectory, 
                                       console.history().size(), 
                                       workingEnvironment->get("BRASH_USER"), 
                                       workingEnvironment->get("BRASH_HOSTNAME")
                                      )
                        );

    
    string userHomeDir = workingEnvironment->get("BRASH_USER_HOME");

    if(!userHomeDir.empty() && !noReadBrashrc)
    {
       FileName rcfile = userHomeDir;

       if(!rcfile.empty() && rcfile[ rcfile.size() - 1] != '/' && rcfile[ rcfile.size() - 1] != '\\')
       {

#ifdef _MSC_VER
           rcfile += '\\';
#else
           rcfile += '/';
#endif
       }

       rcfile += ".brashrc";

       if(!specifiedRCFileName.empty())
       {
          rcfile = specifiedRCFileName;
       }

       if(rcfile.exists())
       {
           cxxtls::StringExpr::ExecutionContext::ArgsVector    args;

           args.push_back(rcfile);

           brashReadAndExecuteInContext(rcfile, myContext.get(), args, false);
       }
       else
       if(!specifiedRCFileName.empty())
       {
           cerr << "User specified started resource file, " << specifiedRCFileName << ", could not be found" << endl;
       }


       string brashNoBannerVariable = myContext->variables().find("BRASH_NO_BANNER");

       if(brashNoBannerVariable.empty())
       {
          string brashBanner = "\r\nWelcome to Brash " BRASH_VERSION_STRING ", a 'bourne shell' clone for MS Windows filesystems\r\n\r\nUse command 'builtins' to get a list of builtin commands\r\n\r\n";


          myContext->fileHandles_->writeBytes(1, brashBanner);
       }

    }

    for(;;)
    {   // read/execute till user quits 

        setConsoleRunFlag(true);

        ScriptInputLine commandString;

        if(!handleLineContinuations(console,
                                    commandString
                                   )
          )
        {
           break;
        }


        if(commandString.text_.empty())
           continue;  // ignore empty lines


        bool wereDone = false;

        string::const_iterator f = commandString.text_.begin(),
                               l = commandString.text_.end();

        EndOfStatementListDetector endDetector;


        while(!wereDone)
        { // split the command line into separate statements
                   

                auto_ptr<cxxtls::ScriptFunctions::StatementList>  statementList(new cxxtls::ScriptFunctions::StatementList);

                auto_ptr<cxxtls::ScriptFunctions::StatementListStatement>  statementListStatement;

                {
                    // keep the following statements together for thread safety

                    statementListStatement.reset( new cxxtls::ScriptFunctions::StatementListStatement(statementList.get()) );
                    statementList.release();
                }

                string filenameAtStart = console.filename();
                int    lineNumberAtStart = console.lineNumber();

                ParseStatementReturnCode rc=
                parseStatement(
                               f,
                               l,
                               console,
                               *statementListStatement,
                               *ScriptThread::ManagedFileHandles::mainStdHandles,
                               commandString,
                               endDetector,
                               myContext.get()
                              );


                if(rc != PSRC_GOOD)
                {
                   console.saveHistory();

                   if(!handleLineContinuations(console,
                                               commandString
                                              )
                     )
                   {
                      break;
                   }

                   f = commandString.text_.begin();
                   l = commandString.text_.end();
           
                   continue;
                }



               if(statementListStatement->statements_->empty())
               {
                   console.saveHistory();

                   if(!handleLineContinuations(console,
                                               commandString
                                              )
                     )
                   {
                      break;
                   }

                   f = commandString.text_.begin();
                   l = commandString.text_.end();

                   continue;
               }
               setConsoleRunFlag(true); // make sure execution is not cancelled

                int jobId = BrashScriptCommandHandler::newJobId();
                
                BrashScriptCommandHandler::registerJob(JobDefinition(jobId, commandString.text_ ) ) ;

                // note:  When you invoke ThreadContext::create() you are getting a fairly thin wrapper around env_ and the file handles
                // that you pass in.  The object so returned is mostly just an "owner" object.  That owner has some functionality but
                // primary storage of environment variables is in the this->env_.  


                auto_ptr<BrashStatementList>  statements;

                statements.reset(new BrashStatementList( static_cast<BrashScriptCommandHandler*>(owner_), // owner
                                                         ScriptThread::ThreadContext::
                                                                 create(myContext->variables().scriptName(),
                                                                        commandLine,
                                                                        workingEnvironment,
                                                                        ScriptThread::ManagedFileHandles::mainStdHandles->clone()
                                                                       ),
                                                         new cxxtls::ScriptFunctions::StatementList,
                                                         jobId,
                                                         true // this is the console!  need this to get the curdir from job exit status
                                                       )
                                );
                

                cxxtls::ScriptThread::ThreadContext *commandContext = statements->context_;

                commandContext->variables().setErrorCode(exitCode);  // copy the exit code from one command to the other
                commandContext->fileHandles_->setCurrentDirectory(currentDirectory);
                commandContext->variables().setLastParmOfLastCommand(myContext->variables().lastParmOfLastCommand());
                commandContext->variables().setGlobbing(myContext->variables().globbing());

                {
                    // keep the following together for exception safety
                    statements->statements_->push_back(ThreadSafeHandle<Statement>(statementListStatement.get()));  
                    statementListStatement.release();
                }

                if(owner_->terminated())
                {
                   wereDone = true;
                   break;
                }

                owner_->push(  statements.release() );   // enque for execution

                // execution is happening now
                
                // the following code only works because we have set the "startedByConsole" flag when constructing the BrashStatementList

                BrashThreadExitType exitInfo = BrashScriptCommandHandler::waitForJob(jobId);

                exitCode = exitInfo.exitCode_;

                if (exitCode == -2)
                {
                    wereDone = true;
                    break;
                }

                currentDirectory = exitInfo.currentDirectory_;

                console.setWorkingDirectory(currentDirectory);

                commandLine = exitInfo.arguments_;

                myContext->variables().setArguments(commandLine);

                myContext->variables().setLastParmOfLastCommand(exitInfo.lastParm_);

                myContext->variables().assign("DIRSTACK", exitInfo.dirStack_);

                myContext->variables().assign("IFS", exitInfo.IFS_);

                myContext->variables().setGlobbing(exitInfo.expandingNames_);

                workingPrompt = workingEnvironment->get("PS1");
                
                if(workingPrompt.empty())
                   workingPrompt = defaultPrompt;

                console.changePrompt(computePrompt(workingPrompt, 
                                                   currentDirectory, 
                                                   console.history().size(),
                                                   workingEnvironment->get("BRASH_USER"),
                                                   workingEnvironment->get("BRASH_HOSTNAME")
                                                  )
                                    );


                while (f != l)
                {
                    if (!isspace(*f))
                        break;
                    ++f;
                }

                if (f == l)
                {
                    if(owner_->terminated())
                    {
                       wereDone=true;
                       break;
                    }

                    console.saveHistory();
                
                    if (!handleLineContinuations(console,
                                                 commandString
                                                )
                        )
                    {
                        break;
                    }
                    f = commandString.text_.begin();
                    l = commandString.text_.end();
                }


                // set the exit code to be rc
            
        }  // split the command line into separate statements

        if(wereDone)
           break;


   

    }  // read/execute till user quits

}
catch(std::exception const &e)
{
    ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2, string("brash FATAL unhandled exception: ") );
    char const *message =  e.what();

    ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2, message, strlen(message) );
    
    ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2, nl);


}
} // run






//==================================================================================================
//==================================================================================================
//==================================================================================================

void
ReadFileCommand::
run()
{
try
{

    cxxtls::StringExpr::ExecutionContext::ArgsVector commandLine;

    AutoDestroyingPointer<cxxtls::ScriptThread::ThreadContext>   myContext(ScriptThread::ThreadContext::create(filename_,
                                                                                                               commandLine,
                                                                                                               &env_,
                                                                                                               ScriptThread::ManagedFileHandles::mainStdHandles->clone()
                                                                                                              )
                                                                          );

    string userHomeDir = env_.get("BRASH_USER_HOME");

    if(!noReadBrashrc && !userHomeDir.empty())
    {
       FileName rcfile = userHomeDir;

       if(!rcfile.empty() && rcfile[ rcfile.size() - 1] != '/' && rcfile[ rcfile.size() - 1] != '\\')
       {

#ifdef _MSC_VER
           rcfile += '\\';
#else
           rcfile += '/';
#endif
       }

       rcfile += ".brashrc";


       if(!specifiedRCFileName.empty())
       {
          rcfile = specifiedRCFileName;
       }

       if(rcfile.exists())
       {
           cxxtls::StringExpr::ExecutionContext::ArgsVector    args;

           args.push_back(".brashrc");

           brashReadAndExecuteInContext(rcfile, myContext.get(), args, false);
       }
       else
       {
           cerr << "Error:  user specified startup resource file, " << specifiedRCFileName << ", could not be found" << endl;
       }

    }

    auto_ptr<ScriptLineStream>   actualStream;

    if(stringStream_ != 0)
    {
       actualStream.reset(  new ScriptLineStream(*stringStream_) );
    }
    else
    if(filename_ == "-")
        actualStream.reset(  new ScriptLineStream("brash>") );
    else
        actualStream.reset(  new ScriptLineStream(filename_, "brash>") );
        

    ScriptLineStream  &inputFile = *actualStream;
    

    if(inputFile.error())
    {
        ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2, inputFile.filename() + colon + lexical_cast<string>(inputFile.lineNumber()) + ":  Error invalid syntax\r\n");
        return;
    }

    int jobId = BrashScriptCommandHandler::newJobId();
    
    BrashScriptCommandHandler::registerJob(JobDefinition(jobId, filename_ ) ) ;   // needs to include arguments,  but do this later


    auto_ptr<BrashStatementList>  statements(new BrashStatementList( static_cast<BrashScriptCommandHandler*>(owner_), // owner
                                                                     ScriptThread::ThreadContext::
                                                                             create(filename_,
                                                                                    args_,
                                                                                    &env_,
                                                                                    ScriptThread::ManagedFileHandles::mainStdHandles->clone()
                                                                                   ),
                                                                     new cxxtls::ScriptFunctions::StatementList,
                                                                     jobId
                                                                   )
                                            );
    
    bool parseOk =parseBrash(inputFile,
                             *statements,
                             *ScriptThread::ManagedFileHandles::mainStdHandles,
                             statements->context_
                            );

    if(!parseOk)
    {
        ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2, inputFile.filename() + colon + lexical_cast<string>(inputFile.lineNumber()) + ":  Error, invalid syntax\r\n");
        return;  // assume someone else explained why its not running
    }

    {
        owner_->push(statements.get());   statements.release();
     }           
    BrashScriptCommandHandler::waitForJob(jobId);


}
catch(std::exception const &e)
{
    ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2, string("brash FATAL unhandled exception: ") );
    char const *message =  e.what();

    ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2, message, strlen(message) );
    
    ScriptThread::ManagedFileHandles::mainStdHandles->writeBytes(2, nl);


}
} // run


BrashThreadExitType BrashThreadExitType::unset(-1);

BrashThreadExitType BrashThreadExitType::shuttingDown(-2);


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

    BourneForStatement::
   ~BourneForStatement()
   {
       CXXTLS_FOREACH(cxxtls::StringExpression *cur, valueList_)
           delete cur;
   }

///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////


    StatementReturnCode 
    BourneForStatement::
    execute(ScriptThread::ThreadContext             *executionContext_in,
            FunctionScope                           *functionScope,
            cxxtls::ScriptFunctions::JobController  *jobController,
            bool                                     skipArg0
           ) const
    {
        // executes both "for" and "select" statements


        AutoDestroyingPointer<cxxtls::ScriptThread::ThreadContext>       clonedContext;

        ScriptThread::ThreadContext *executionContext = executionContext_in;

        if(redirections_.get() && !redirections_->empty())
        {
            clonedContext.reset( executionContext_in->cloneForSameJob(executionContext_in->variables().scriptName(), executionContext_in->variables().arguments()) );

            executionContext = clonedContext.get();

            RedirType::redirectFileHandles(executionContext, redirections_.get());
        }


        executionContext->fileHandles_->setNoTerminatePipe(0);  // don't cancel upstream writes when commands with io redirection terminate
       
       
       std::list<std::string> strings;

       // parse the expressions and split them using the IFS variable

       if(!valueList_.empty())
       {
           CXXTLS_FOREACH(cxxtls::StringExpression *cur, valueList_)
           {
               StringArray value;
           
               cur->executeHelper(&executionContext->variables(), value, true, true);
           
               // now we have the evaluated value -- we must now split based on IFS
           
               for(size_t i = 0; i < value.size(); ++i)
               {
                   std::string const &s = value[i];
           
                   strings.push_back(s);
               }
           }
       }
       else
       {
          // if there is no user supplied list, then use $@

          int count=0;

          CXXTLS_FOREACH(string const &cur, executionContext_in->variables().arguments())
          {
              if(++count == 1)
                 continue;


              strings.push_back(cur);
          }

       }

       // we now have all the for loop variable strings over which to iterate.

       if(isSelect_)
       {
         for(;;)
         {
             // handle the "select" execution
         
               if(!consoleStillRunning())
                  throw ScriptThread::UserIOCancellation();

             vector<string const *>  items;  items.reserve(strings.size());
         
             size_t count = 0;
         
             string ps3 = executionContext->variables().find("PS3");
         
             if(ps3.empty())
                ps3 = string("Enter a number in the range 1 - ") + lexical_cast<string>(strings.size()) + string(": ");
         
             // write the data to be selected from
         
             CXXTLS_FOREACH(std::string const &cur, strings)
             {
                items.push_back(&cur);
         
                executionContext->fileHandles_->writeBytes(1, lexical_cast<string>(items.size()) + string(":  ") + cur + nl);
                
             }
         
             // read from the user
         
             executionContext->fileHandles_->writeBytes(2, ps3);
         
             // let the user answer the question
         
             string inputLine;
         
             if(!executionContext->fileHandles_->readLine(0, inputLine) )
             {
                executionContext->fileHandles_->writeBytes(2, string("select-statement:  expected a number in the range of 1 to ") + lexical_cast<string>(items.size()) + nl);
                executionContext->variables().setErrorCode(1);
                continue;
                
             }
         
             // interpret the user's results
         
             int value =-1;
         
             sscanf(inputLine.c_str(), "%d", &value);
         
             if(value < 1 || value > int(items.size()))
             {
                executionContext->fileHandles_->writeBytes(2, string("select-statement:  expected a number in the range of 1 to ") + lexical_cast<string>(items.size()) + nl);
                executionContext->variables().setErrorCode(1);
                continue;
                 
             }
         
             executionContext->variables().assign(varname_, *items[value-1] );
            
             // loop forever on this statement until the user makes a selection that executes code which performs a break; statement.
         
         
                 StatementReturnCode rc = loopStatement_->execute(executionContext, functionScope, jobController, skipArg0);
                 
                 if(rc == SRC_Break)
                    break; // but don't return rc == break

                 if(rc != SRC_Success && rc != SRC_Continue)
                 {
                    return rc;
                 }

                 // ignore continue since that is what we are doing anyway
         
         }
       }
       else
       if(isArithmetic_ && strings.size() == 3)
       {
           // silently act wierd if we don't have the correct count of strings.  the parser should never let us get here
           // if the syntax of the for loop is wrong.

           std::list<std::string>::const_iterator a = strings.begin();

           std::string loopVar;
           std::string loopCondition;
           std::string loopIncrement;

           std::string tmp1, tmp2, tmp3;


           try
           {
               loopVar        = expandStringVariableReferences(tmp1 = *a++, executionContext);
               loopCondition  = expandStringVariableReferences(tmp2 = *a++, executionContext);
               loopIncrement  = expandStringVariableReferences(tmp3 = *a++, executionContext);
           }
           catch(std::exception const &e)
           {
                executionContext->fileHandles_->writeBytes(2, string("brash for-statement loop conditions invalid:  ((") +
                                                              tmp1 + semicolon + tmp2 + semicolon + tmp3 + "))\r\n");
                executionContext->variables().setErrorCode(1);

                return SRC_Success;
               
           }



           // perform the initial assignment if there is one

            string::const_iterator f = loopVar.begin(), l = loopVar.end();

            std::auto_ptr<StringExpr::Arithmetic::Assignment> assignment(StringExpr::Arithmetic::parseAssignment(f,l));

            if(assignment.get())
                assignment->execute(&executionContext->variables());


           for(;;)
           {
               if(!consoleStillRunning())
                  throw ScriptThread::UserIOCancellation();

               f = loopCondition.begin();
               l = loopCondition.end();

               std::auto_ptr<StringExpr::Arithmetic::Assignment> condition(StringExpr::Arithmetic::parseAssignment(f,l));

               if(condition.get())
                  if(!condition->execute(&executionContext->variables()))
                      break;

           
               StatementReturnCode rc = loopStatement_->execute(executionContext, functionScope, jobController, skipArg0);
           
               if(rc == SRC_Break)
                   break;
               else
               if(rc == SRC_Continue)
                   continue;
               else
               if (rc == SRC_Success)
               {
                   f = loopIncrement.begin();
                   l = loopIncrement.end();

                   std::auto_ptr<StringExpr::Arithmetic::Assignment> increment(StringExpr::Arithmetic::parseAssignment(f, l));

                   if (increment.get())
                       increment->execute(&executionContext->variables());

               
                   continue;
               }
               else
                  return rc;
           
              
           }

       }
       else
       {
           // handle the "for" execution
           CXXTLS_FOREACH(std::string const &cur, strings)
           {
               if(!consoleStillRunning())
                  throw ScriptThread::UserIOCancellation();

               executionContext->variables().assign(varname_, cur);
           
               StatementReturnCode rc = loopStatement_->execute(executionContext, functionScope, jobController, skipArg0);
           
               if(rc == SRC_Break)
                   break;
               else
               if(rc == SRC_Continue)
                   continue;
               else
               if(rc == SRC_Success)
                   continue;
               else
                  return rc;
           
           
           
           }
       }

       return SRC_Success;  // IF NOTHING GOES WRONG...
        
    }


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

StatementReturnCode 
BourneWhileStatement::
execute( ScriptThread::ThreadContext             *executionContext_in,
         FunctionScope                           *functionScope,
         cxxtls::ScriptFunctions::JobController  *jobController,
         bool                                     skipArg0
       ) 
       const
{
        AutoDestroyingPointer<cxxtls::ScriptThread::ThreadContext>       clonedContext;

        ScriptThread::ThreadContext *executionContext = executionContext_in;

        CXXTLS_HEAVYDEBUG("WhileStatement begins");

        if(redirections_.get() && !redirections_->empty())
        {
            clonedContext.reset( executionContext_in->cloneForSameJob(executionContext_in->variables().scriptName(), executionContext_in->variables().arguments()) );

            executionContext = clonedContext.get();

            RedirType::redirectFileHandles(executionContext, redirections_.get());
        }


        executionContext->fileHandles_->setNoTerminatePipe(0);  // don't cancel upstream writes when commands with io redirection terminate

    for(;;)
    {
        if(!consoleStillRunning())
        {
           throw ScriptThread::UserIOCancellation();
        }

        if(executionContext->loggingExecution())
        {
           executionContext->fileHandles_->writeBytes(2, string("+ while\r\n"));
        }

        CXXTLS_HEAVYDEBUG("WhileStatement executing condition");

        StatementReturnCode rc = condition_->execute(executionContext, functionScope, jobController, skipArg0);

        if(rc != SRC_Success)
        {
           CXXTLS_HEAVYDEBUG("WhileStatement condition clause bad rc  = " << int(rc));

           return rc;
        }

        if(executionContext->variables().errorCode() != 0)
        {
           CXXTLS_HEAVYDEBUG("WhileStatement loop termination");

           break;
        }
        // while clause is true, so lets execute the statement body

        if(executionContext->loggingExecution())
        {
           executionContext->fileHandles_->writeBytes(2, string("+ while-do\r\n") );
        }
  
        CXXTLS_HEAVYDEBUG("WhileStatement executing loopBody");

        rc = loopBody_->execute(executionContext, functionScope, jobController, skipArg0);

        if(rc == SRC_Break)
        {
           CXXTLS_HEAVYDEBUG("WhileStatement loop body executed 'break' ");

           break;  // but don't return rc == break!
        }

        if(rc != SRC_Success && rc != SRC_Continue)
        {
           CXXTLS_HEAVYDEBUG("WhileStatement body clause clause failed, rc = " << int(rc));

           return rc;
        }

        CXXTLS_HEAVYDEBUG("WhileStatement after loopBody");


        // don't care about its exit code

    }

        CXXTLS_HEAVYDEBUG("WhileStatement ends");


    return SRC_Success;


}

    BourneWhileStatement::
   ~BourneWhileStatement()
   {
   }


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

StatementReturnCode 
BourneIfStatement::
execute( ScriptThread::ThreadContext            *executionContext_in,
         FunctionScope                          *functionScope,
         cxxtls::ScriptFunctions::JobController *jobController,
         bool                                    skipArg0
       ) 
       const
{
        AutoDestroyingPointer<cxxtls::ScriptThread::ThreadContext>       clonedContext;

        ScriptThread::ThreadContext *executionContext = executionContext_in;


        if(redirections_.get() && !redirections_->empty())
        {
            clonedContext.reset( executionContext_in->cloneForSameJob(executionContext_in->variables().scriptName(), executionContext_in->variables().arguments()) );

            executionContext = clonedContext.get();

            RedirType::redirectFileHandles(executionContext, redirections_.get());
        }
       
       

    {

        if(executionContext->loggingExecution())
        {
           executionContext->fileHandles_->writeBytes(2, "+ if\r\n", 6);
        }

        StatementReturnCode rc = condition_->execute(executionContext, functionScope, jobController, skipArg0);

        if(rc != SRC_Success)
           return rc;

        if(executionContext->variables().errorCode() == 0)
        {
            // if clause is true, so lets execute the statement body
            
            if(mainClause_)
            {
         
                if(executionContext->loggingExecution())
                {
                   executionContext->fileHandles_->writeBytes(2, string("+ if-then-clause\r\n") );
                }
        
                rc = mainClause_->execute(executionContext, functionScope, jobController, skipArg0);
                
                if(rc != SRC_Success)
                   return rc;
            }
        }
        else
        if(elseClause_)
        {
            if(executionContext->loggingExecution())
            {
               executionContext->fileHandles_->writeBytes(2, string("+ if-else-clause\r\n") );
            }
            rc = elseClause_->execute(executionContext, functionScope, jobController, skipArg0);
            
            if(rc != SRC_Success)
               return rc;
        }

    }

    return SRC_Success;


}

    BourneIfStatement::
   ~BourneIfStatement()
   {
   }


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

BournePipeStatement::
~BournePipeStatement()
{
   CXXTLS_DEBUG("~BournePipeStatement(" << std::hex << (long )(this) << ")");  
}

StatementReturnCode 
BournePipeStatement::
execute(ScriptThread::ThreadContext             *executionContext_in,
        FunctionScope                           *functionScope,
        cxxtls::ScriptFunctions::JobController  *jobController,
        bool                                     skipArg0
       ) const
{
    // first, clone the execution context to make sure that the file handles of interest are copies of the current context's file handles.
    //
    // Of course, the variables are the same
    
    AutoDestroyingPointer<ScriptThread::ThreadContext>  clonedContext( executionContext_in->cloneForSameJob(executionContext_in->variables().scriptName(), executionContext_in->variables().arguments()) );

    ScriptThread::ThreadContext* executionContext = clonedContext.get();

    // increment pipe depth

    string curDepth = executionContext->variables().find("BRASH_PIPE_DEPTH");

    if(curDepth.empty())
    {
       executionContext->variables().assign("BRASH_PIPE_DEPTH", "1");
    }
    else
    try
    {
        int depth = lexical_cast<int>(curDepth);

        ++depth;

        executionContext->variables().assign("BRASH_PIPE_DEPTH", lexical_cast<string>(depth));
    }
    catch(...)
    {
       executionContext->variables().assign("BRASH_PIPE_DEPTH", "10");
    }

    // In reverse order, create new "jobs" for each element of the chain by pushing them
    // into the jobController (ie BrashScriptCommandHandler) input queue.
    //
    // Then,  wait for all jobs to finish.  Return the error code of the last job.
    //
    // We must create 

    if(chain_.empty())
    {
        executionContext_in->variables().setErrorCode(1);
        throw EmptyPipeChainException();
    }

    size_t threadCount = lexical_cast<size_t>(executionContext->variables().find("BRASH_THREAD_COUNT"));

    if(threadCount > 2 && (chain_.size() > threadCount -2 ) )
    {
        executionContext->variables().assign("BRASH_PIPE_DEPTH", curDepth);
        executionContext_in->variables().setErrorCode(1);
       throw PipeChainTooLongException("Chain Member Count", (chain_.size()), "Allowed Size", (threadCount-2));
    }


    BrashScriptCommandHandler *owner = static_cast< BrashScriptCommandHandler * >(jobController);

    std::list< ThreadSafeHandle<Statement> >::iterator f = chain_.begin(),
                                                       l = chain_.end();

    ScriptThread::ThreadContext *  contextToUse = executionContext;

    ScriptThread::ThreadContext *  nextContextToUse=0;

    // give the current context to the end of the pipe chain,
    // then create a context for each of the pipe predecessors in sequence.
    // Each of which will have it stdout bound to the stdin in of the 
    // downstream statement.

    int waitForJob = -1;

    bool deleteContext = false;

    ThreadGate finishLine(chain_.size());  // all pipe components need to finish before we can be done

    bool closeFileHandles = false;  // only false the first time through

    // before creating the tail end of the pipe, lets set the "ignoreTerminationRequest" flag on its stdout file handle because the
    // tail is shared by other statements after the pipe and we don't want to shut off i/o to them when the pipe terminates. 

    ScriptThread::ManagedFileHandles::FileHandleAttributes tailStdout = contextToUse->fileHandles_->getFileAttributes(1);

    if(tailStdout.pipeState_)
    {
       tailStdout.pipeState_->ignoreTerminationRequest();
       
    }


    try
    {
        int reverseChainIndex = 0;

        while(f != l)
        {
            --l;


            bool terminateInputPipe = (f != l); // don't do it at the beginning

            ThreadSafeHandle<Statement> &s = *l;
    
            if(l != f)
            {
               // we are not at the begining of the chain, so we need to clone the predecessor
               // file handles for the predecessor to write to in order to get data to the current
               // job

               CXXTLS_LOG("PipeStatement:  " << (void*)(this) << " Creating Piped Context for next Member");
    
               nextContextToUse  =   contextToUse->pipe();
            }
            else
            {
               CXXTLS_LOG("PipeStatement:  " << (void*)(this) << " Creating the last Member");
               nextContextToUse = 0; // clear out the pointer
            }
    
            // at this point, we want to create a job that runs the current statement.
            // except for the first such job, being the end of the pipe chain, the
            // job should delete its copy of the context passed to it.  The end of the
            // chain is using our context and should not.
            //

            int jobId = BrashScriptCommandHandler::newJobId();
                
            BrashScriptCommandHandler::registerJob(JobDefinition(jobId, "pipemember" ) ) ;

            if(waitForJob < 0)
               waitForJob = jobId;

            auto_ptr<ScriptCommand>  command(new BrashPipeMember(
                                                                  owner,
                                                                  contextToUse,
                                                                  s,
                                                                  jobId,
                                                                  deleteContext,
                                                                  finishLine,
                                                                  reverseChainIndex,
                                                                  closeFileHandles,
                                                                  terminateInputPipe
                                                                )
                                            );

             closeFileHandles = true;

             ++reverseChainIndex;

            {
                // keep together for exception saftey

                CXXTLS_DEBUG("BournePipeStatement(" << std::hex << (long )(this) << ") starting pipe member (" << (long)(command.get()) << ")" );  

                owner->push(  command.get() );   // enque for execution
                command.release();
            }

            Sleep(1); // give the child a chance to start executing. 
#if 0
            Sleep(50); // give the child a longer chance
                       // this is necessary because when grep is piped to sed, and the input is small, grep terminates before sed has a 
                       // change to fully initialize.  However, this is a band aid.  What we need is to wait until sed is fully up and
                       // running before allowing grep to run.  This delay will work most of the time, but if the machine is heavily
                       // loaded, sed make take longer than 50 milliseconds to initialize and thus band aid won't work.  And it slows things
                       // down for everyone all the time -- unnecessarily if the downstream pipe item is not an executable.
#endif
    
            deleteContext = true;
            contextToUse = nextContextToUse;
        }
    }
    catch(...)
    {
       ScriptThread::ThreadContext::destroy(nextContextToUse);
       executionContext->variables().assign("BRASH_PIPE_DEPTH", curDepth);
        executionContext_in->variables().setErrorCode(1);
       throw;
    }


    BrashThreadExitType exitInfo = BrashScriptCommandHandler::waitForJob(waitForJob);


    finishLine.wait(); // just wait until all threads have signaled completion.

    executionContext->variables().assign("BRASH_PIPE_DEPTH", curDepth);

    return SRC_Success;
}

StatementReturnCode 
BrashScriptCommandHandler::
executeDetached(ScriptThread::ThreadContext *executionContext,
                FunctionScope                *functionScope,
                JobController                *jobController,
                ThreadSafeHandle<Statement>  &statement
               ) 
{

    auto_ptr<cxxtls::ScriptFunctions::StatementList>  statementList(new cxxtls::ScriptFunctions::StatementList);

    statementList->setAlreadyDetached(true);

    statementList->push_back(statement);

    auto_ptr<cxxtls::ScriptFunctions::StatementListStatement>  statementListStatement;

    {
        // keep the following statements together for thread safety

        statementListStatement.reset( new cxxtls::ScriptFunctions::StatementListStatement(statementList.get()) );
        statementList.release();
    }


    int jobId = BrashScriptCommandHandler::newJobId();


    string commandName="[brash][unknown]";


    BrashScriptCommandHandler::registerJob(JobDefinition(jobId, executionContext->variables().scriptName() ) ) ;

    // note:  When you invoke ThreadContext::create() you are getting a fairly thin wrapper around env_ and the file handles
    // that you pass in.  The object so returned is mostly just an "owner" object.  That owner has some functionality but
    // primary storage of environment variables is in the this->env_.  


    auto_ptr<BrashStatementList>  statements;

    statements.reset(new BrashStatementList( this,
                                             executionContext->cloneForNewJob(),
                                             new cxxtls::ScriptFunctions::StatementList,
                                             jobId,
                                             false // not launched from the console
                                           )
                    );

    {
        // keep the following together for exception safety
        statements->statements_->push_back(ThreadSafeHandle<Statement>(statementListStatement.get()));  
        statementListStatement.release();
    }


    {
        // keep together for exception saftey
        push(  statements.get() );   // enque for execution
        statements.release();
    }

    return SRC_Success;
}


InstallFunctionStatement::~InstallFunctionStatement() {}

StatementReturnCode 
InstallFunctionStatement::
execute(ScriptThread::ThreadContext             *executionContext,
        FunctionScope                           *functionScope,
        ScriptFunctions::JobController          *jobController,
        bool                                     skipArg0
       ) const
{
   executionContext->variables().variables().addFunction(name_, const_cast<ThreadSafeHandle<Statement>&>(body_), file_, line_);
   // no effect on the current error code
    return SRC_Success;
}


BourneCaseStatement::~BourneCaseStatement() {}

StatementReturnCode 
BourneCaseStatement::
execute(ScriptThread::ThreadContext             *executionContext_in,
        FunctionScope                           *functionScope,
        cxxtls::ScriptFunctions::JobController  *jobController,
        bool                                     skipArg0
       ) const
{

    AutoDestroyingPointer<cxxtls::ScriptThread::ThreadContext>       clonedContext;

    ScriptThread::ThreadContext *executionContext = executionContext_in;

    if(redirections_.get() && !redirections_->empty())
    {
        clonedContext.reset( executionContext_in->cloneForSameJob(executionContext_in->variables().scriptName(), executionContext_in->variables().arguments()) );

        executionContext = clonedContext.get();

        RedirType::redirectFileHandles(executionContext, redirections_.get());
    }

    FileName testValue;

    StringArray output;

    testExpression_->execute(&executionContext->variables(), output);

    testValue = output.combine();

    if(executionContext->loggingExecution())
    {
        string logMessage("+ case [");

        logMessage += testValue;
        logMessage += "] in\r\n";

        executionContext->fileHandles_->writeBytes(2, logMessage);

    }

    // ok, we now have a value to compare against.  Let us iterate over the cases an evaluate the
    // their individual conditions (i.e file name patterns) match the testValue.  We will stop when
    // we find the first one, and then we will execute its statements and quit.

    if(cases_.get() == 0 || cases_->empty())
    {
       return SRC_Success;   // nothing to do, I suppose.  This should never happen.  The parser should prevent this
                             // could throw an exception.
    }

    CaseMemberList const &cases = *cases_.get();



    CXXTLS_FOREACH(CaseMemberList::value_type const &cur, cases)
    {
       CXXTLS_FOREACH(StringExpression * subPattern, *cur->conditions_)
       {
           string pattern;
           
           output.clear();
           
           subPattern->executeHelper(&executionContext->variables(), output, false, false);   // do not glob
           
           pattern = output.combine();

           bool usePatternMatch = false;

           // check to see if we should do a simple string compare or pattern match.
           // if even one node in the comparison pattern contains *, ?, or [, then it is a 
           // pattern match

           std::list<StringExpr::Node*>::iterator f = subPattern->begin(),
                                                  l = subPattern->end();

           while(f != l)
           {
               StringExpr::Node* node = *f++;

               if( dynamic_cast<StringExpr::GlobNode*>(node))
               {
                  usePatternMatch = true;
                  break;
               }
           }

           bool foundMatch = usePatternMatch ? testValue.matches(pattern.c_str())
                                             : pattern == testValue
                                             ;

           
           if(foundMatch)
           {
               if(executionContext->loggingExecution())
               {
                   string logMessage("+ case member (");
           
                   logMessage += pattern ;
                   logMessage += ") matches\r\n";
           
                   executionContext->fileHandles_->writeBytes(2, logMessage);
           
               }

               return cur->statements_->execute(executionContext, functionScope, jobController, skipArg0);
           }
       }
    }

    return SRC_Success;
}

NotStatement::~NotStatement()
{
}


StatementReturnCode 
NotStatement::
execute(ScriptThread::ThreadContext             *executionContext_in,
        FunctionScope                           *functionScope,
        cxxtls::ScriptFunctions::JobController  *jobController,
        bool                                     skipArg0
       ) const
{


    StatementReturnCode rc = mainClause_->execute(executionContext_in,
                                                  functionScope,
                                                  jobController,
                                                  skipArg0
                                                 );

    executionContext_in->variables().setErrorCode(!executionContext_in->variables().errorCode());

    if(executionContext_in->loggingExecution())
    {
        executionContext_in->fileHandles_->writeBytes(2, string("+ ! -- inverting previous exit code\r\n"));
    }


    return SRC_Success;
}

BourneConditionalStatement::~BourneConditionalStatement()
{
}


StatementReturnCode 
BourneConditionalStatement::
execute(ScriptThread::ThreadContext             *executionContext_in,
        FunctionScope                           *functionScope,
        cxxtls::ScriptFunctions::JobController  *jobController,
        bool                                     skipArg0
       ) const
{

    StatementReturnCode rc = 
                             lhs_->execute(executionContext_in,
                                           functionScope,
                                           jobController,
                                           skipArg0
                                          );

    if(rc != SRC_Success)
    {
        return rc;
    }

    if(isAndAnd_)
    {
        if(0 == executionContext_in->variables().errorCode() )
        {
           return  
                  rhs_->execute(executionContext_in,
                                functionScope,
                                jobController,
                                skipArg0
                               );
        }
    }
    else
    {
        if(0 != executionContext_in->variables().errorCode() )
        {
           return  
                  rhs_->execute(executionContext_in,
                                functionScope,
                                jobController,
                                skipArg0
                               );
        }
    }

    return SRC_Success;
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

StatementReturnCode 
BourneBlockStatement::
execute( ScriptThread::ThreadContext            *executionContext_in,
         FunctionScope                          *functionScope,
         cxxtls::ScriptFunctions::JobController *jobController,
         bool                                    skipArg0
       ) 
       const
{
    AutoDestroyingPointer<cxxtls::ScriptThread::ThreadContext>       clonedContext;

    ScriptThread::ThreadContext *executionContext = executionContext_in;

    if(redirections_.get() && !redirections_->empty())
    {
        clonedContext.reset( executionContext_in->cloneForSameJob(executionContext_in->variables().scriptName(), executionContext_in->variables().arguments()) );

        executionContext = clonedContext.get();

        RedirType::redirectFileHandles(executionContext, redirections_.get());
    }

    
    ScriptThread::ManagedFileHandles::iterator stdinInfo = executionContext->fileHandles_->find(0);

    bool stdinExists = false;

    if(stdinInfo != executionContext->fileHandles_->end())
        stdinExists = true;

    bool savedNoTerminatePipe = false;

    if(stdinExists)
    {
        savedNoTerminatePipe = !executionContext->fileHandles_->find(0)->second.terminatePipe_;
        executionContext->fileHandles_->setNoTerminatePipe(0, true);
    }

    StatementReturnCode rc;

    try
    {
       rc = block_->execute(executionContext, functionScope, jobController, skipArg0);

       if(stdinExists)
            executionContext->fileHandles_->setNoTerminatePipe(0, savedNoTerminatePipe);

    }
    catch(...)
    {
       if(stdinExists)
           executionContext->fileHandles_->setNoTerminatePipe(0, savedNoTerminatePipe);

       throw;
    }

    return rc;

}

    BourneBlockStatement::
   ~BourneBlockStatement()
   {
   }



BrashWriteExpressions::~BrashWriteExpressions()
{
   CXXTLS_FOREACH(StringExpression *cur, expressions_)
       delete cur;
}

        

StatementReturnCode 
BrashWriteExpressions::
execute(ScriptThread::ThreadContext             *executionContext,
        FunctionScope                           *functionScope,
        cxxtls::ScriptFunctions::JobController  *jobController,
        bool                                     skipArg0
       ) const
{
    CXXTLS_FOREACH(StringExpression const *cur, expressions_)
    {
        string value;

        StringArray output;

        cur->execute(&executionContext->variables(), output);

        value = output.combine();

        executionContext->fileHandles_->writeBytes(1, value + nl);
    }

    return SRC_Success;
}



#ifdef _MSC_VER
   // DOH, SPOS MICROSOFT CRAP.  the mechanism for taking control of a process tree
   // is so poorly documented, that have had to come up with a temporary work around
   // for just dealing with SIGBREAK, SIGKILL, etc.
   //
   // the work around is to have to processes if you are launching a console window:
   // the first process is a non-Console-aware process.  When you start a console
   // version of brash, it first, sets up a SIGINT handler, then launches brash as a 
   // child process from that one.  The brash child process then becomes windows-console
   // aware.  It is only through this redirection that proper catching of ^C but also the
   // killing of child proceses when it is caught actually works.
   //
   // I should rewrite the console reader logic such that it it doesn't need to do this,
   // and when I do, this goofiness will go away.

   //
   //  Whenever any command executes in the console, the following functions are used to 
   //  save environment.
   //

    std::string environmentEncode(std::string const &s)
    {
       std::string rv;

       rv.reserve(s.size());

       CXXTLS_FOREACH(char c, s)
       {
           if(c == '\n')
           {
               rv += '\\';
               rv += 'n';
           }
           else
           if(c == '\\')
           {
              rv += '\\';
              rv += '\\';
           }
           else
              rv += c;
       }

       return rv;

    }

    std::string environmentDecode(std::string const &s)
    {
       std::string rv;

       rv.reserve(s.size());

       string::const_iterator f = s.begin(), l = s.end();

       while(f != l)
       {
           char c = *f++;

           if(c == '\\')
           {
              if(f == l)
                 rv += '\\';
              else
              {
                 if(*f == 'n')
                 {
                    rv += nl;
                    ++f;
                 }
                 else
                    rv += *f++;
              }
           }
           else
              rv += c;
       }

       return rv;

    }



    std::string loadSavedEnvironment(std::string const &loadFile, StringExpr::ExecutionVariables &recoveredEnvironment, std::string const &ownerId)
    {
        // return the current directory in the rv but the recovertedEnvironment through the parameter.

        std::string curdir;

        ifstream file(loadFile.c_str());

        std::string line;

        if(!file.good())
        {
           return recoveredEnvironment.get("OLDPWD");
        }

        if(!::getline(file, curdir))
        {
           cerr << "brash:  Fatal error:  unable to read to file '" << loadFile << "' in loadSavedEnvironment()" << endl;
           exit(1);
        }

        FileName tmp(curdir);

        if(!tmp.is_dir())
        {
           cerr << "brash:  Fatal error: loaded current directory is invalid '" << curdir << "' in loadSavedEnvironment()" << endl;
           exit(1);
        }

        while(::getline(file, line))
        {
            std::string prep = environmentDecode(line);

            size_t equal;

            if( (equal = prep.find_first_of('=') ) >= prep.size())
            {
               cerr << "brash:  Fatal error: missing '=' in variable assignment from file '" << loadFile << "' in loadSavedEnvironment()" << endl;
               exit(1);
            }

            std::string name = prep.substr(0,equal);

            recoveredEnvironment.set(name, prep.substr(equal+1));

        }

        return curdir;
    }
    void saveEnvironment(std::string const &saveFile, StringExpr::ExecutionVariables &env, std::string const &ownerId, std::string const &currentDirectory)
    {
        ofstream file(saveFile.c_str());
        
        if(!file.good())
        {
           cerr << "brash:  Fatal error:  unable to write to file '" << saveFile << "' in saveEnvironment()" << endl;
           exit(1);
        }

        file << currentDirectory << endl;

        cxxtls::StringExpr::ExecutionVariables::const_iterator f = env.begin(), l = env.end();
   
          while(f != l)
          {
              cxxtls::StringExpr::ExecutionVariables::VariablesMap::value_type const &cur = *f++;

              file << cur.first;
              file << '=';
              file << environmentEncode(cur.second.value());
              file << nl;
          }
    }




#endif

std::pair<std::string, bool> findBrashExecutableInPath(std::string const &name, std::string const &PATH)
{

  StrTool::stringlist_t splitpath;

#ifdef _MSC_VER
  StrTool::Is_Delim functor(';');
  char slash = '\\';
#else
  StrTool::Is_Delim functor(':');
  char slash = '/';
#endif

  StrTool::parse_words(PATH,
                       &splitpath,
                       2000000000U,
                       functor
                      );

  StrTool::stringlist_t::iterator f = splitpath.begin();
  StrTool::stringlist_t::iterator l = splitpath.end();

  string extension;

  size_t lastDot = name.find_last_of('.');

  if(lastDot < name.size())
  {
     extension = StrTool::lowercase(name.substr(lastDot));  // includes the dot.
  }

  FileName commandNamePath(name);

  if (commandNamePath.exists())
  {
     // we probably have a full or relative pathname to a script

      ifstream x(commandNamePath.c_str());

      string line;

      std::getline(x, line, char(0));

      if (line.size() > 2 && line[0] == '#' && line[1] == '!')
      {
          return std::pair<std::string, bool>(commandNamePath, true);  // call this a winner
      }

  }

  while(f != l)
  {
      std::string curDir = *f++;

      StrTool::trim(curDir);

      FileName tmp(curDir);

      if(!tmp.is_dir())
         continue;       // ignore missing directories

      curDir = tmp;

      tmp += slash;
      tmp += name;

      if(tmp.exists())
      {
          if(extension == ".sh" || extension == ".brash")
          {
             return std::pair<std::string,bool>(tmp,true); // we have a winner
          }
          else
          if(extension.empty())
          {
              ifstream x(tmp.c_str());

              string line;

              std::getline(x, line, char(0));

              if(line.size() > 2 && line[0] == '#' && line[1] == '!')
              {
                  return std::pair<std::string,bool>(tmp,true);  // call this a winner
              }

          }
          else
          {
             // ignore unknown file types
          }
      }
      else
      {
#ifdef _MSC_VER
          // now check MSDOS PATHEXT

          char const *pathext = getenv("PATHEXT");

          if(pathext)
          {
              StrTool::stringlist_t exetypes;

              StrTool::parse_words(pathext,
                                   &exetypes,
                                   2000000000U,
                                   functor
                                  );
            
              StrTool::stringlist_t::iterator a = exetypes.begin(),
                                              b = exetypes.end();
              while(a != b)
              {
                  string cur = tmp + *a++;

                  FileName curFile(cur);

                  if(curFile.exists())
                      return std::pair<std::string,bool>(cur,false);
              }
          }
#endif
      }
  }


    return std::pair<std::string,bool>("",false);
}

static string expandStringVariableReferences(string const &s, cxxtls::ScriptThread::ThreadContext *context)
{
    string rv;

    string::const_iterator f = s.begin(),
                           l = s.end();


    while(f != l)
    {
       // parse and execute string fragments

       char c = *f;

       switch(c)
       {
          default:  rv += c;
                    ++f;
                    break;

          case '$':
          case '"':
          case '\'':
          case '`':
                  string::const_iterator exprStart = f;

                  auto_ptr<StringExpression> expression( StringExpr::parseStrExpr(f,l, false) );

                  if(expression.get())
                  {
                     StringArray output;

                     expression->execute(&context->variables(), output);

                     rv += output.combine();

                  }
                  else
                  {
                      rv += c;
                      f = exprStart;
                      ++f;
                  }
                  break;
       }


    }

    return rv;


}

