//
// Copyright 2002,2015 Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

#ifndef CUT_HEADER_INCLUDED_P
#define CUT_HEADER_INCLUDED_P

#include <string>
#include <list>
#include <vector>
#include <utility>

namespace cxxtls

{

class CutSet
  //.  A CutSet object defines a range of characters to be extracted from
  //.  a string.  The range is a list of pairs of character position numbers (1 based).
{
  typedef std::pair<int,int>   RangeType;  // first to last (inclusive)
                                           // 0 based numbers!
  typedef std::list<RangeType> RangeList;

  RangeList                    ranges_;   // 0 based numbers

  int maxPos_;  // maximum char pos (1 based)

public:

  bool empty() const { return ranges_.empty(); }
    //.  Return true if this cutset is empty (ie no ranges defined)


  CutSet(std::string const &); 
    //.  construct a CutSet given a string of this form:
    //.
    //.    "1-5,10,11,20-45"
    //.
    //.  When using the fectCharacters() function:
    //.
    //.  @li  this means:  characters 1-5, also characters 10 and 11, and finally
    //.       characters 20-45.
    //.
    //.  But when using fetchDelimitedFields():
    //.
    //.  @li  the definition means:  field 1-5, 10, 11, and 20 through 45.
    //.       where the fields are separated by the delimiter character.
    //.
    //.  And when using fetchWords, the definition of the cutset is the same
    //.  as when using fetchDelimitedfields(), except that the multiple
    //.  whitespace characters serve as a single delimiter.
    //.
    //.  NOTE for convenience to the user, the cut set is expressed in 1
    //.  based numbers, but internally, 0 based numbers are used
    //.
    //.  Note that if a string is shorter than the cutset range expects, nothing
    //.  will included in the output for that invalid range -- thus you get
    //.  safe failure modes.
    //.
    //.  Note also that instea

  std::string fetchCharacters (std::string const &input) const;
    //.  Splits the specified input string into separate characters and returns
    //.  those characters defined by the constructor or by the use of the 
    //.  addCharPos() or addRange() functions.

  std::string fetchDelimitedFields (std::string const &input_string, char delim) const;
    //.  Splits the input_string on each instance of the specified delimiter character and returns a
    //.  a string which is the concatenation of all the fields.  The delimiter character
    //.  separates the subfields within the returned string.
    //.
    //.  Note that if the set is specified using the addRange() and/or the addCharPos()
    //.  functions, then the cutset does not define character positions but rather
    //.  substring indices.  

  std::string fetchWords (std::string const &input_string) const;
    //.  Splits the input_string on whitespace between words and returns a
    //.  a string which is the concatenation of all the words separated by a single space each.
    //.
    //.  Note that if the set is specified using the addRange() and/or the addCharPos()
    //.  functions, then the cutset does not define character positions but rather
    //.  substring indices.  
    //.
    //.  See Strtool::parse_words() for a definition of what whitespace characters are.


  void addRange(int first, int last) 
    { 
      --first;  // convert 1 based parameters to 0 based internal rep form.
      --last;

      ranges_.push_back(std::pair<int,int>(first, last)); 

      if(last > maxPos_)
        maxPos_ = last;
    }
    //.  Add a range of character positions to the selection set.

  void addCharPos(int pos) { addRange( pos, pos ); }
    //.  Adds a single character range to the selection set

  std::string fetchFields(std::vector<std::string> const &fields, char delim) const;
    //.  Given an array of fields, use *this as a definition of which fields are of interest and return
    //.  a string containing the specified fields.  The specified delimiter will be used to delimit
    //.  the fields in the return string -- unless the delimiter is 0, in which case, no delimiter
    //.  will be used.


};  // class CutSet

}  // namespace cxxtls

#endif
