#ifndef patternmatch_header_included_p
#define patternmatch_header_included_p
//
// Copyright 2012, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

#include <string>
#include <vector>
#include <utility>

namespace cxxtls
{

struct Charset;  // forward declaration;

struct PatternMatcher
   ///  An object that lets you match file name patterns, just like you would do in
   ///  bash.  This is not a regex style pattern but a "glob" style pattern match.
   ///
   ///  This class exists to be incorporated into script language interpreters.
{
    typedef std::string::const_iterator const_iterator;   // input source text iterator

    struct Matcher;  // forward declaration of the pattern matcher expression
                     // terms.

    enum MatchType
    {
        EXACT,                     // like normal bash glob operations
        SHORTEST_LEADING,          // like the bash expression:  ${f#<exprexssion>}
        LONGEST_LEADING,           // like the bash expression:  ${f##<exprexssion>}
        SHORTEST_TRAILING,         // like the bash expression:  ${f%<exprexssion>}
        LONGEST_TRAILING           // like the bash expression:  ${f%%<exprexssion>}
    };

    typedef std::pair<size_t,size_t> OffsetAndLength;


private:


    typedef std::vector<Matcher*> MatchExpression;

    MatchExpression  matchExpression_;

    void parsePattern(std::string const &pattern);
        ///< Convert a pattern definining string into a vector of Matcher*'s which can
        ///  then be executed in sequence to compare a given string to the pattern.
        ///  Example strings are:
        ///
        ///    "abc.txt"
        ///    "*.cpp"
        ///    "a*b

    size_t leadingMatch(MatchType t, std::string const &comparand) const
        // a helper used by shortest and longest leading match
    {
         const_iterator first = comparand.begin(), last = comparand.end();

         return leadingMatch(t, first, last);
    }

    size_t leadingMatch(MatchType, const_iterator &nc, const_iterator &lc) const; 
        // a helper used by shortest and longest leading match

    size_t trailingMatch(MatchType, std::string const &comparand) const; 
        // a helper used by shortest and longest trailing match

public:

    PatternMatcher(std::string const &pattern)
    {
       parsePattern(pattern);   // convert the pattern string into terms in the
                                // matchExpression_;
    }

    ~PatternMatcher();


    bool matches(std::string const &comparand) const;
        ///   Returns  true if and only if the comparand exactly matches the
        ///   the pattern.

    size_t shortestLeadingMatch(std::string const &comparand) const;
        ///<  Returns the length of the shortest leading match of the pattern
        ///   match as applied to comparand.  If the full pattern does not match
        ///   the leading characters in comparand, zero will be returned.
        ///   that is no partial matches will be returned.
        ///
        ///   Only if there are wildcard characters in the pattern is there
        ///   any doubt about the length of the matched pattern.  when the
        ///   shortest leading match is requested for a pattern that ends in a
        ///   wildcard, the wildcard at the end is effectively ignored.
        ///
        ///   If you get any non-zero value then the comparand does begin
        ///   with 1 instance of the pattern.
        ///
        ///   For example, if the pattern is 'abc', then comparand must begin
        ///   with 'abc', or zero will be returned.  
        ///
        ///   If the pattern is 'a*bx', then the return value will be non-zero
        ///   if the comparand begins with 'a' and has at least 1 'bx' in it.
        ///   The return value will be the length from the beginning of the
        ///   string to that first bx.

    size_t longestLeadingMatch(std::string const &comparand) const;
        ///<  Returns the length of the longest leading match of the pattern
        ///   match as applied to comparand.  If the full pattern does not match
        ///   the leading characters in comparand, zero will be returned.
        ///   that is no partial matches will be returned.
        ///
        ///   Only if there are wildcard characters in the pattern is there
        ///   any doubt about the length of the matched pattern.
        ///
        ///   If you get any non-zero value then the comparand does begin
        ///   with 1 instance of the pattern.
        ///
        ///   For example, if the pattern is 'abc', then comparand must begin
        ///   with 'abc', or zero will be returned.  
        ///
        ///   If the pattern is 'a*bx', then the return value will be non-zero
        ///   if the comparand begins with 'a' and has at least 1 'bx' in it.
        ///   The return value will be the length from the beginning of the
        ///   string to the last bx foundin the string.  Consider this example:
        ///
        ///       v-----v                   shortest match
        ///       a12345bx98876abxuvw.      comparand string
        ///       ^-------------^           longest match


    size_t shortestTrailingMatch(std::string const &comparand) const;
        ///<  Returns the length of the shortestTrailing match of the pattern
        ///   against the the specified comparand.  For example, if the pattern
        ///   is 'a' and the comparand does not end in 'a', then the return value
        ///   is 0 -- meaning that there is no match.   If the comparand ends in 
        ///   'a' then the return value is one.
        ///
        ///   Only if there are wildcard characters in the pattern is there
        ///   any doubt about the length of the matched pattern.
        ///
        ///   But if the pattern is 'a*', then the return value will be length
        ///   of the string at includes the last 'a' and all the remaining 
        ///   characters in the comparand.
        ///
        ///   If the pattern is '*.c', then return value is equal to the length
        ///   of the whole string if the string ends in '.c'.  Otherwise, the
        ///   return value will be zero.
        ///
        ///   If the pattern is "/*.c", and if the comparand ends in "/stuff.c",
        ///   then the length will be 8 (the length of /stuff.c).

    size_t longestTrailingMatch(std::string const &comparand) const;
        ///<  Returns the length of the longest trailing match of the pattern
        ///   against the specified comparand.  If there are no matches, then
        ///   zero is returned.  Only if the pattern contains wildcard characters
        ///   is there in doubt about the length of the matched pattern.
        ///
        ///   If the pattern is "/*", then the returned match length would be
        ///   from the first / to the end of the string -- assuming there was
        ///   a / in the comparand, of course.

    OffsetAndLength  find(std::string const &s, size_t startPos=0) const;
        ///<  Return the offset from the beginning of string s to the
        ///   first character of a region of characters that match this pattern.
        ///   The startPos parameter creates a bias such that the actual scanning
        ///   starts at beginning+startPos.
        ///   If no match is found, return std::string::npos;

    std::string removeLeadingMatch(std::string const &s, bool longestMatch=false) const;
        ///<  Return a copy of the input string, s.  If this pattern matches
        ///   the beginning of the string, then the copy will be modified
        ///   such that the leading match will be removed.  If the longestMatch
        ///   flag is specified, then the longest leading match will be removed.
        ///   If no match is found, return std::string::npos;

    std::string removeTrailingMatch(std::string const &s, bool longestMatch=false) const;
        ///<  Return a copy of the input string, s.  If this pattern matches
        ///   the end of the string, then the copy will be modified
        ///   such that the trailing match will be removed.  If the longestMatch
        ///   flag is specified, then the longest trailing match will be removed.
        ///   If no match is found, return std::string::npos;

   std::string replaceMatches(std::string const &s, std::string const &r, bool allMatches=false) const;
        ///<  Return a copy of s with the first or all matches of this pattern
        ///   replaced by the value in r.  If the allMatches flag is true then
        ///   all matches will be replaced, not just the first.
        ///   If no match is found, return std::string::npos;

private:
    void addLiteralTerm(std::string &tmp);  // just a helper

};

};  // namespace cxxtls


#endif
