#ifndef Sequence_Map_h_included
#define Sequence_Map_h_included


//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.




///@file
///
///  This header file defines the Sequence_Map class.  A Sequence_Map is a
///  collection of related sequences of object.  A Sequence_Map helps you
///  determine whether you have a sequence which matches one of the patterns
///  in the tree.  See Sequence_Map.

#include <cxxtls/hash_list.h>
#include <iostream>
#include <stdlib.h>

namespace cxxtls
{

template<class T, class ID, class H = Hasher<T> >
class Sequence_Map
/// 
///  A Sequence_Map is a collection of related sequences of T objects.
///  Sequences supported by a Sequence_Map can not contain optional members.
///  The template argument, H is the name of a class object with an
///  operator()(T) that produces and integer value which is a 'hash' of T.
/// 
///  Each Sequence has a given name, specified by the ID class.  One
///  purpose of a sequence map is to provide you with a mechanism for
///  analyzing sequences of T's to determine which named sequence
///  it maches, if any.  That is, you define a bunch of sequence of T's.
///  You identify each sequence by giving it a unique ID.  You can then
///  compare any random sequence to see if it has a match in the Sequence_Map
///  and if so, what is its ID.
/// 
///  You can also use the Sequence_Map to aid in parsing random sequences
///  and bomb out early when you realize that you don't have a valid sequence.
///  A Sequence_Map can help you implement syntax directed parsing.
///  For example, suppose you are parsing output from a computer program and
///  you want to detect and act upon certain strings but not all.  The emacs
///  gud-mode does this.  A sequence map can help with this.  Another use is
///  in parsing function keys from terminals.  Sometimes unix buffering
///  causes function key sequences to be split across multiple 'read()' calls.
///  You can't guarantee that you'll get all 5 characters in the f12 sequence,
///  for example, in a single read() operation.  Yet you don't want to hard
///  read for 5 characters all the time because the key sequences are often
///  shorter than 5 characters -- typically they are 1, 3, or 5 characters
///  long, but not always.
/// 
///  The design for Sequence_Map is roughly based on a Huffman coding
///  scheme -- although Huffman would probably spin in his grave if he
///  saw how big and complex this mechanism is -- he was interested in
///  compressing bit patterns for com networks.  Here we are interested
///  in a simple, universally applicable interface that provides often
///  need tools for simple pattern recognition.
/// 
///  The algorithmic complexity of a Sequence_Map is as follows:  Once the
///  sequences are inserted into the map, searching the map for a given
///  item in the sequence is O(1).  That is, if have N items in your test
///  sequence, it will take roughly O(N) steps to validate whether or not it
///  is a good sequence -- and get its ID value.  Inserting a new sequence
///  into the map is approximately O(N) where N is the length of the sequence.
/// 
///  Here's how you use a sequence map:
/// 
///  @li  Define some sequences, and give each a unique ID.
/// 
///  @li Iterate over each element of a test sequence to see if it is the
///    Sequence_Map.
/// 
///  To perform a search, you ask the Sequence_Map for an iterator like
///  object which contains the context of a search.  You can then iterate
///  over your test sequence, one item at a time, and ask the context object
///  if this is a member of any valid sequence in the sequence map.  It will
///  give you the following information:
/// 
///  @li No, this T is not part of a valid sequence
///  @li Yes, this T is part of a sequence and it in does in fact terminate
///      the sequence whose ID is XXX.
///  @li Yes this T is part of a sequence, but it is not a complete sequence.
/// 
///  Here is an example:
///  @code
///     struct H { int operator() (int t) { return t; } };  // int hasher
/// 
///     typedef Sequence_Map<char,int,H> SEQMAP;
///                                       // or accept the default hasher which is
///                                       // Hasher<char>
///     typedef SM::sequence             SEQUENCE;
///     typedef SM::search               SEARCH;
/// 
///     SEQMAP tree;
/// 
///       SEQUENCE s1(tree, 1); // fred
///                s1 += 'f';   s1 += 'r';   s1 += 'e' ;   s1 += 'd';
///                if(! s1.complete() )
///                   error("subsequence of longer sequence");
/// 
///       SEQUENCE s2(tree, 2); // frat
///                s2 += 'f';   s2 += 'r';   s2 += 'a' ;   s2 += 't';
///                if(! s2.complete() )
///                  error("subsequence of longer sequence");
/// 
///       // tree now has 2 sequences in it.
/// 
///       int  id;
///       bool found = tree.find("frat", 4, &id);  // id is 2, found is true
/// 
///       char const *search="fred";
/// 
///       char const *scan = search;
/// 
///       SEARCH c(tree);
/// 
///       while(*scan != 0 && c(*scan))
///       {
///          ++scan;
///          ++c;
///       }
/// 
///       if(*scan == 0 && c.finished(&id) )
///       {
///         // id will have 1 in it because the context, c, indicates
///         // that the subsequence has been completed.
///       }
///   @endcode
///   Notes:
/// 
///   @li Before calling finished(), you must have incremented the
///     search object after having called it in order to determine
///     if a value is part of the sequence -- and having it return true
///     to indicate yes.
/// 
///   @li The example above is only a comment, see the Sequence_Map::find
///     method below for real working code.
/// 
/// 
{
public:

  //
  //  A sequence map implements a tree of nodes.  The path from the root of the tree
  //  to any specific node defines a sequence.  Complete sequences should end in nodes
  //  whose 'next' pointer is null.  Each node in the tree is an entry in a hash_list
  //  which defines elements which are a pair<T, node*>.  That is, the T is the 'first'
  //  member of the pair.  The node* is the 'second' member of the pair and it
  //  refers to the 'next' step in the sequence.
  //
  //  Each complete sequence -- ie one which ends in a node whose 'second' member is
  //  null, appears as an entry in hash_list<node*, ID>.  That is, determining the
  //  ID of a given complete sequence envolves looking up the final node in the
  //  id map to get the id value.
  //


  typedef Sequence_Map<T,ID,H> self;    ///< helper typedef that refers to the sequence map

  struct                       node;    ///< description of the rest of a given sequence
					///  assuming you are already in the sequence
					///  somewhere
				
  typedef hash_list<T,node*,H> rest_type; ///< data structure representing the rest of a
					  ///  all possible sequences after the current
					  ///  sub-sequence of T's.
					
  struct node
  : public rest_type
  {
    // this class MUST CONTAIN NOTHING but a rest_type!
  };

  struct H1
  {
    int operator() (node* n) 
    { 
      union   
      {
        node *n1;
        int   i1;
      } junk;

      junk.n1 = n;

      return junk.i1;
    }
  };


  typedef hash_list<node*,ID,H1> id_map_type;  ///< storage type for complete sequence ID's.

  Sequence_Map() {}

    ///< Construct an empty sequence map

 ~Sequence_Map() { destroy_tree(); }

    ///< The _destructor_!  What will they think of next!

  bool find(T const *first, size_t count, ID *id)

    /// Compare an array to see if it is complete sequence and get the
    /// sequence id.
    ///
  {
     if(count == 0)
       return false;

    search c(*this);

    while(count && c(*first))
    {
      ++first;
      --count;
      ++c;
    }

    if(count != 0)
      return false;

    if(c.finished(id))
      return true;

    return false;  // did not find an exact match.

  }

  template<class Iterator>
  bool find(Iterator first, Iterator last, ID *id)

    /// Compare a range to see if is an exact match for an named pattern.
    /// If so, return true and the id of the matched pattern.
    /// If not return false, and don't change the id.

  {
    search c(*this);

    while(first != last && c(*first))
    {
      ++first;
      ++c;
    }

    if(first != last)
      return false;

    if(c.finished(id))
      return true;

    return false;  // did not find an exact match.
  }

  class search

    /// A 'search' is a context for comparing an arbitrary sequence
    /// with those stored in the Sequence_Map.  The idea of course is
    /// to determine whether or not a given sequence exactly matches
    /// one found in the map -- and also what the ID of that sequence
    /// is.  A search is a kind of iterator that scans through the
    /// structure of the Sequence_Map.  You have to call operator++
    /// to get to the next location in the sequence.  If you want to
    /// know if you are sitting at the end of the sequence, call
    /// finished.  A search does not exactly model an iterator so
    /// it has no comparison functions, and the Sequence_Map has
    /// no begin or end methods.
  {
    self*      tree_;  ///< which tree are we talking about
    node*      scan_;  ///< where are we currently in the tree
    node*      next_;  ///< next location to goto in the tree when operator++ is called

  public:

    search(self &tree)
    : tree_(&tree),
      scan_((node*)(&tree.root_)),
      next_(0)
    {
      ///  Construct a search on a given sequence map -- and starting
      ///  at the top level of its tree of sequence data.
    }

    bool operator() (T const & value)
    {
      /// Check to see if a specified value is a valid next member of the
      /// sequence up to this point.  Sets the next_ member.

      // scan_ points to a hash_list of pair<T,node*>;  If the specified
      // T is the key value of any of these pairs then the specified T is
      // a valid member of the current sequence.
      //

      next_ = 0;

      if(scan_->size() == 0)
	return false;
	
      typename rest_type::iterator where = scan_->find(value);

      if(where == scan_->end())
	return false;
	
      next_ = where->data_.second;

      return true;

    }

    void operator++()
    {
      /// Assuming that someone has already called operator()(T const&),
      /// and have gotten an true indication, step to the next item
      /// in the sequence along this direction.
      ///
      /// Note that multiple calls to operator() (T const&) can be
      /// made.  The most recent call is the one that defines the direction
      /// operator++ will take -- if any

      if(next_ == 0)
      {
	std::cerr << "Error, can't increment search till you match an item" << std::endl;
	std::cerr << "       see " << __FILE__ << ":" << __LINE__ << std::endl;
	exit(1);
      }

      scan_ = next_;
      next_ = 0;

    }

    bool finished(ID* id)
    {
      /// Determine if the current location in the sequence is the end of
      /// a sequence.  A true return value means yes and the id will be
      /// updated.  Otherwise false will be returned and the id will not
      /// be updated.

      if(scan_->size() == 0)
      {
	typename id_map_type::iterator f = tree_->ids_.find(scan_);
	
	if(f != tree_->ids_.end())
	{
	  *id = f->data_.second;
	  return true;
	}
	
      }
	
      return false;

    }
  };


  class sequence

    /// A sequence is a handle object that lets you create patterns in the tree
    /// You construct a sequence object given a Sequence_Map object.  You then
    /// add members to the sequence.  When you are done, you call the
    /// complete() function to see if this is a valid (complete) sequence
    /// or if you have mistakenly create a subsequence of an existing
    /// sequence.
  {
  public:
    typedef Sequence_Map<T,ID,H> seq_map;

  private:

    seq_map*   tree_;   ///< which Sequence_Map are we talking about
    ID         id_;     ///< what sequence are we creating
    node* scan_;        ///< where in the sequence map are we currently sitting


  public:

    sequence(seq_map& tree, ID const &v)
    :  id_(v),
       scan_((node*)(&tree.root_))
    {
      /// Create a handle to a sequence in a specified tree

      tree_ = &tree;
    }


   ~sequence()
   {
      /// Destroy the sequence handle, but not the sequence
   }

    bool complete() const
    {
      /// Assert that the sequence actually is complete -- ie that you
      /// have not specified a new sequence which is a subset of a
      /// smaller bigger sequence.  This function should only be called when
      /// feel yourself to be complete with a sequence because it both
      /// performs the test AND marks the sequence as complete.  Once a
      /// sequence is marked as complete, you can't add a new longer sequence
      /// with the same initial elements as an existing complete sequence --
      /// nor can you create a new shorter sequence.  The return code indicates
      /// that you have successfully created the sequence.  A return value of false
      /// means that the sequence is empty or that it is a subsequence of a bigger
      /// sequence.

      if(scan_->size() != 0)
	return false;

      tree_->ids_[scan_] = id_;

      return true;
    }

    void operator+= (T const &new_item)
    {

      /// Extend the sequence with a new item.  Warning:  this is not meant
      /// to be a particularly robust interface.  If you attempt to extend
      /// an existing complete sequence, you'll get a fatal error.  The goal
      /// here is to provide a workable interface.  If you want to make sure
      /// that you don't accidentally crash the program, -- say when users
      /// supply the sequence -- you should call is_complete() to see if
      /// you if the current sequence is already complete before adding an
      /// item to it -- and getting a crash.

      typename id_map_type::iterator where = tree_->ids_.find(scan_);

      if(where != tree_->ids_.end())
      {
	 std::cerr << "Error, you can't extend a complete sequence" << std::endl;
	 std::cerr << "       see " << __FILE__ << ":" << __LINE__ << std::endl;
	 exit(1);
      }

      typename rest_type::iterator f = scan_->find(new_item),
				   e = scan_->end();
			
      if(f != e)
      {
	// we are still matching part of an existing sequence
	
	scan_ = f->data_.second;
	
      }
      else
      {
	// adding a new tail to an existing sequence

	node* next_node = (node*)(new rest_type);
	
	typedef typename rest_type::mapped_type entry_type;

	scan_->insert( entry_type(new_item, next_node) );

	scan_ = next_node;
      }

    }

    bool can_complete() const
    {
      /// Check to see if this sequence is complete at its current size.  To
      /// complete a sequence you must call 'complete()' and the sequence
      /// must be unique.

      if(scan_->size() != 0)
	return false;

      if(scan_ == &tree_->root_)
	return false;
	
      return true;
    }

  };


private:

  friend class sequence;
  friend class search;

  rest_type    root_;   ///< the list of T's that begin sequences
  id_map_type  ids_;    ///< list of all sequences in this sequence map

  void destroy_tree()
    /// private method that recursively frees up all memory used by the tree.
  {
     typename rest_type::iterator f = root_.begin(),
				  l = root_.end();
			
     while(f != l)
     {
       typename rest_type::mapped_type &r = f->data_;

       node* n = r.second;

       delete n;

       ++f;
	
     }
  }

};

} // namespace cxxtls


#endif
