#include <string>
#include <vector>

namespace cxxtls
{

struct StringArray
    ///  An Array of strings.
    ///
    ///  If you attempt to read a non-existant value,
    ///  you will get an empty value.
    ///
    ///  If you attempt to write to a non-existent value,
    ///  you will create empty strings up to the index
    ///  that you are writing -- unless you crash, of course..
    ///
{

    StringArray()
    {
    }

    ~StringArray()
    {
       std::vector<std::string*>::iterator first = data_.begin(),
                                           last  = data_.end();
       while(first != last)
       {
          delete *first++;
       }
    }

    size_t size() const { return data_.size(); }

    std::string operator[] (size_t index)  const
        ///   Return the value at location index or an empty string
        ///   if said location does not exist.
    {
        if(index < data_.size())
            return *data_[index];

        return std::string();
       
    }

    struct indexWrapper;

    friend class indexWrapper;

    struct indexWrapper
        /// When you use operator[] of a mutable StringArray,
        /// you actually get one of these objects which can automatically
        /// convert itself to the string you are requesting.
        ///
        /// Thus const vectors are faster to access than non-const.
        ///
        /// This data type makes sure that when you index a particular
        /// location, it actually exists.  It does this by creating
        /// needed items.
        ///
        /// Note that this is not meant as a general purpose class.  There
        /// is no reason to declare a variable of this type.  Instead,
        /// you should only create this as a result of using the array
        /// index operator of a mutable StringArray.
    {
        std::vector<std::string*> &vector_;
        size_t                     index_;

        indexWrapper(std::vector<std::string*> &data, size_t index)
        :  vector_(data)
        ,  index_(index)
        {
        }

        void ensureSize()
        {
            size_t curSize = vector_.size();

            while(curSize <= index_)
            {
                vector_.push_back( new std::string() );
                ++curSize;
            }

        }

        operator std::string &()
        {
            ensureSize();
            
            // at this point, the vector size is at least 1
            // greater than the item being requested

            return *vector_[index_];
                   
        }


        std::string &operator=(std::string const &source)
        {
            ensureSize();
            
            // at this point, the vector size is at least 1
            // greater than the item being requested

            return (*vector_[index_] = source);
        }

        std::string &operator=(char const *source)
        {
            ensureSize();
            
            // at this point, the vector size is at least 1
            // greater than the item being requested

            return (*vector_[index_] = source);
        }



        std::string const *valuePtr() const
            /// return a pointer to either a null string or
            /// the actual data associated with this indexWrapper
        {
           static std::string null;

           std::string *valuePtr;

           if(index_ < vector_.size())
              valuePtr = vector_[index_];
           else
              valuePtr = &null;

           return valuePtr;
        }

        bool operator== (char const *rhs) const { return (*valuePtr()) == rhs; }
        bool operator!= (char const *rhs) const { return (*valuePtr()) != rhs; }
        bool operator>= (char const *rhs) const { return (*valuePtr()) >= rhs; }
        bool operator<= (char const *rhs) const { return (*valuePtr()) <= rhs; }
        bool operator>  (char const *rhs) const { return (*valuePtr()) >  rhs; }
        bool operator<  (char const *rhs) const { return (*valuePtr()) <  rhs; }

        bool operator== (std::string const &rhs) const { return (*valuePtr()) == rhs; }
        bool operator!= (std::string const &rhs) const { return (*valuePtr()) != rhs; }
        bool operator>= (std::string const &rhs) const { return (*valuePtr()) >= rhs; }
        bool operator<= (std::string const &rhs) const { return (*valuePtr()) <= rhs; }
        bool operator>  (std::string const &rhs) const { return (*valuePtr()) >  rhs; }
        bool operator<  (std::string const &rhs) const { return (*valuePtr()) <  rhs; }


        std::string &operator+= (char const *rhs)
        {
           ensureSize();

           return *vector_[index_] += rhs;
        }

        std::string &operator+= (char rhs)
        {
           ensureSize();

           return *vector_[index_] += rhs;
        }


        std::string &operator+= (std::string const &rhs)
        {
           ensureSize();

           return *vector_[index_] += rhs;
        }

        template<class Stream>
        friend Stream &operator<< (Stream &s, indexWrapper const &me)
        {
           if(me.index_ < me.vector_.size())
               return s << *me.vector_[me.index_];

           return s;
        }


    };  // struct indexWrapper

    indexWrapper operator[] (size_t index)
    {
        return indexWrapper(data_, index);
    }


    void push_back(std::string const &v)
    {
       data_.push_back( new std::string(v) );
    }

    indexWrapper back()
    {
       return indexWrapper(data_, data_.empty() ? 0 : data_.size()  -1);
    }

    void pop_back()
    {
      delete data_.back();
      data_.pop_back();
    }

    std::string combine() const
        /// concatenate all members into one string, separated
        /// by spaces.
    {

        std::string rv;

        std::vector< std::string * >::const_iterator first = data_.begin(),
                                                     last  = data_.end();

        while(first != last)
        {
            std::string *cur = *first++;

            rv += *cur;

            if(first != last && !rv.empty())
            {
              rv += ' ';
            }
            
        }

        return rv;

    }

    void clear() 
    { 
        for(size_t i = 0; i < data_.size(); ++i)
            delete data_[i];
           
        data_.clear(); 
    }

    bool empty() { return data_.empty(); }

    void swap(StringArray &other) throw ()
        /// safely swap the contents of this string array.
    {
       data_.swap(other.data_);
    }

    void erase(size_t i)
    {
       if(i < data_.size())
       {
          data_.erase(data_.begin() + i);
       }
    }


private:

    std::vector< std::string* > data_;

    StringArray(StringArray const &);     // disallow copy construct
    void operator=(StringArray const &);  // disallow assignment

}; // class StringArray

} // namespace cxxtls
